import math
import re

import FreeCAD

from ..utils import q_form as q_form
from ..utils.basic_functions_part1 import is_opposite, is_parallel
from .string_functions import remove_redundant


class CardLine:
    def __init__(self, card, linesize=80, tabspace=6, fmt=""):
        self.str = card + " "
        self.lineSize = linesize
        self.__leftspace__ = linesize - len(self.str)
        self.tabspace = tabspace
        self.tabstr = " " * tabspace
        self.fmt = fmt

    def extend(self, dataList):
        line = ""
        for data in dataList:
            data_str = "{:{fmt}} ".format(data, fmt=self.fmt)
            if self.__leftspace__ - len(data_str) > 0:
                line += data_str
                self.__leftspace__ -= len(data_str)
            else:
                self.str += line + "\n"
                line = f"{self.tabstr}{data_str}"
                self.__leftspace__ = self.lineSize - len(line)

        self.str += line


class CellString:
    def __init__(self, linesize=80, tabspace=6):
        self.str = ""
        self.lineSize = linesize
        self.__leftspace__ = linesize
        self.tabspace = tabspace
        self.tabstr = " " * tabspace

    def add(self, string):
        self.str += string

    def wrap_line(self, offset=0):

        self.str = self.str.strip()
        self.str = re.sub(" +", " ", self.str)
        self.str = re.sub(" *\( *", "(", self.str)
        self.str = re.sub(" *\) *", ")", self.str)
        self.str = re.sub(" *: *", ":", self.str)
        self.str = re.sub("\)\(", ") (", self.str)
        self.str = re.sub("(?P<num>\d)\(", "\g<num> (", self.str)
        self.str = re.sub("\)(?P<num>\d)", ") \g<num>", self.str)
        self.str = re.sub("\)-", ") -", self.str)

        if len(self.str) + offset <= self.lineSize:
            return

        newline = ""
        init = 0
        icount = self.lineSize - 1 - offset
        lenleft = len(self.str)

        while True:
            while self.str[icount] not in (")", ":", " "):
                icount -= 1

            newline += self.str[init : icount + 1]
            lenleft -= icount - init + 1
            init = icount + 1
            icount += self.lineSize - self.tabspace
            if lenleft == 0:
                break
            elif icount < len(self.str):
                newline += f"\n{self.tabstr}"
            else:
                newline += f"\n{self.tabstr}{self.str[init:]}"
                break

        self.str = newline


def change_surf_sign(surf, Seq):
    if Seq.level == 0:
        for i, e in enumerate(Seq.elements):
            if surf == abs(e):
                Seq.elements[i] = -Seq.elements[i]
    else:
        for i, e in enumerate(Seq.elements):
            if type(e) is int:
                if surf == abs(e):
                    Seq.elements[i] = -Seq.elements[i]
            else:
                change_surf_sign(surf, e)


def write_mcnp_cell_def(definition, tabspace=0, offset=0):
    sdef = CellString(tabspace=tabspace)
    str_def = remove_redundant(write_sequence_mcnp(definition))
    sdef.add(str_def)
    sdef.wrap_line(offset)
    return sdef.str


def write_serpent_cell_def(definition, tabspace=0, offset=0):
    sdef = CellString(tabspace=tabspace)
    str_def = remove_redundant(write_sequence_serpent(definition))
    sdef.add(str_def)
    sdef.wrap_line(offset)
    return sdef.str


def write_phits_cell_def(definition, tabspace=0, offset=0):
    sdef = CellString(tabspace=tabspace)
    str_def = remove_redundant(write_sequence_phits(definition))
    sdef.add(str_def)
    sdef.wrap_line(offset)
    return sdef.str


def write_openmc_region(definition, options, w_type="XML"):
    if w_type == "XML":
        return write_sequence_omc_xml(definition)
    if w_type == "PY":
        return write_sequence_omc_py(definition, options)


def write_sequence_mcnp(Seq):
    if Seq.level == 0:
        if Seq.operator == "AND":
            line = f"({' '.join(map(str, Seq.elements))})"
        else:
            line = f"({':'.join(map(str, Seq.elements))})"
    else:
        terms = []
        for e in Seq.elements:
            if type(e) is int:
                terms.append(str(e))
            else:
                terms.append(write_sequence_mcnp(e))

        if Seq.operator == "AND":
            line = f"({' '.join(terms)})"
        else:
            line = f"({':'.join(terms)})"

    return line


def write_sequence_serpent(seq):
    if seq.level == 0:
        if seq.operator == "AND":
            line = f"({' '.join(map(str, seq.elements))})"
        else:
            line = f"({':'.join(map(str, seq.elements))})"
    else:
        terms = []
        for e in seq.elements:
            if type(e) is int:
                terms.append(str(e))
            else:
                terms.append(write_sequence_mcnp(e))

        if seq.operator == "AND":
            line = f"({' '.join(terms)})"
        else:
            line = f"({':'.join(terms)})"

    return line


def write_sequence_phits(seq):
    if seq.level == 0:
        if seq.operator == "AND":
            line = f"({' '.join(map(str, seq.elements))})"
        else:
            line = f"({':'.join(map(str, seq.elements))})"
    else:
        terms = []
        for e in seq.elements:
            if type(e) is int:
                terms.append(str(e))
            else:
                terms.append(write_sequence_phits(e))

        if seq.operator == "AND":
            line = f"({' '.join(terms)})"
        else:
            line = f"({':'.join(terms)})"

    return line


def write_sequence_omc_xml(seq):
    if seq.level == 0:
        if seq.operator == "AND":
            line = f"({' '.join(map(str, seq.elements))})"
        else:
            line = f"({' | '.join(map(str, seq.elements))})"
    else:
        terms = []
        for e in seq.elements:
            if type(e) is int:
                terms.append(str(e))
            else:
                terms.append(write_sequence_omc_xml(e))

        if seq.operator == "AND":
            line = f"({' '.join(terms)})"
        else:
            line = f"({' | '.join(terms)})"
    return line


def write_sequence_omc_py(seq, options, prefix="S"):

    strSurf = lambda surf: (f"-{prefix}{-surf}" if surf < 0 else f"+{prefix}{surf}")

    if seq.level == 0:
        if seq.operator == "AND":
            line = f"({' & '.join(map(strSurf, seq.elements))})"
        else:
            line = f"({' | '.join(map(strSurf, seq.elements))})"
    else:
        terms = []
        for e in seq.elements:
            if type(e) is int:
                terms.append(strSurf(e))
            else:
                terms.append(write_sequence_omc_py(e, options))

        if seq.operator == "AND":
            line = f"({' & '.join(terms)})"
        else:
            line = f"({' | '.join(terms)})"
    return line


def mcnp_surface(id, Type, surf, options, tolerances, numeric_format):
    mcnp_def = ""

    if Type == "Plane":
        if surf.pointDef and options.prnt3PPlane:
            P1 = surf.Points[0]
            P2 = surf.Points[1]
            P3 = surf.Points[2]
            mcnp_def = """{:<6d} P   {P1[0]:{d}} {P1[1]:{d}} {P1[2]:{d}} 
           {P2[0]:{d}} {P2[1]:{d}} {P2[2]:{d}}
           {P3[0]:{d}} {P3[1]:{d}} {P3[2]:{d}}""".format(
                id, P1=P1 / 10, P2=P2 / 10, P3=P3 / 10, d=numeric_format.P_d
            )
        else:
            A = surf.Axis.x
            B = surf.Axis.y
            C = surf.Axis.z
            if surf.Axis.isEqual(FreeCAD.Vector(1, 0, 0), tolerances.pln_angle):
                mcnp_def = "{:<6d} PX  {:{x}}".format(id, surf.Position.x / 10.0, x=numeric_format.P_xyz)
            elif surf.Axis.isEqual(FreeCAD.Vector(0, 1, 0), tolerances.pln_angle):
                mcnp_def = "{:<6d} PY  {:{y}}".format(id, surf.Position.y / 10.0, y=numeric_format.P_xyz)
            elif surf.Axis.isEqual(FreeCAD.Vector(0, 0, 1), tolerances.pln_angle):
                mcnp_def = "{:<6d} PZ  {:{z}}".format(id, surf.Position.z / 10.0, z=numeric_format.P_xyz)
            else:
                D = surf.Axis.dot(surf.Position)
                mcnp_def = "{:<6d} P   {:{abc}} {:{abc}} {:{abc}} {:{d}}".format(
                    id,
                    A,
                    B,
                    C,
                    D / 10.0,
                    abc=numeric_format.P_abc,
                    d=numeric_format.P_d,
                )

    elif Type == "Cylinder":
        Dir = FreeCAD.Vector(surf.Axis)
        Dir.normalize()
        Pos = surf.Center * 0.1
        rad = surf.Radius * 0.1
        if is_parallel(Dir, FreeCAD.Vector(1, 0, 0), tolerances.angle):
            if Pos.y == 0.0 and Pos.z == 0.0:
                mcnp_def = "{:<6d} CX  {:{r}}".format(id, rad, r=numeric_format.C_r)
            else:
                mcnp_def = "{:<6d} C/X  {:{yz}} {:{yz}} {:{r}}".format(
                    id, Pos.y, Pos.z, rad, yz=numeric_format.C_xyz, r=numeric_format.C_r
                )
        elif is_parallel(Dir, FreeCAD.Vector(0, 1, 0), tolerances.angle):
            if Pos.x == 0.0 and Pos.z == 0.0:
                mcnp_def = "{:<6d} CY  {:{r}}".format(id, rad, r=numeric_format.C_r)
            else:
                mcnp_def = "{:<6d} C/Y  {:{xz}} {:{xz}} {:{r}}".format(
                    id, Pos.x, Pos.z, rad, xz=numeric_format.C_xyz, r=numeric_format.C_r
                )
        elif is_parallel(Dir, FreeCAD.Vector(0, 0, 1), tolerances.angle):
            if Pos.y == 0.0 and Pos.x == 0.0:
                mcnp_def = "{:<6d} CZ  {:{r}}".format(id, rad, r=numeric_format.C_r)
            else:
                mcnp_def = "{:<6d} C/Z  {:{xy}} {:{xy}} {:{r}}".format(
                    id, Pos.x, Pos.y, rad, xy=numeric_format.C_xyz, r=numeric_format.C_r
                )
        else:
            # Is not still working fine
            Q = q_form.q_form_cyl(Dir, Pos, rad)
            mcnp_def = """\
{:<6d} GQ  {v[0]:{aTof}} {v[1]:{aTof}} {v[2]:{aTof}}
          {v[3]:{aTof}} {v[4]:{aTof}} {v[5]:{aTof}}
          {v[6]:{gToi}} {v[7]:{gToi}} {v[8]:{gToi}}
          {v[9]:{j}} """.format(
                id,
                v=Q,
                aTof=numeric_format.GQ_1to6,
                gToi=numeric_format.GQ_7to9,
                j=numeric_format.GQ_10,
            )

        # Si se quiere rcc en vez de Q form
        #    pnt = surf.Center.sub(surf.Axis.multiply(1.0e7)) # mas alla de 100 m
        #    dir = surf.Axis.multiply(1.0e8)
        #    Vx= pnt.x/10.0
        #    Vy= pnt.y/10.0
        #    Vz= pnt.z/10.0
        #    Hx= dir.x/10.0
        #    Hy= dir.y/10.0
        #    Hz= dir.z/10.0
        #    rad=surf.Radius/10.0
        #    mcnp_def='%i  RCC  %13.7E %13.7E %13.7E %13.7E\n       %13.7E %13.7E %13.7E' %(id,Vx,Vy,Vz,Hx,Hy,Hz,rad)

    elif Type == "Cone":
        Apex = surf.Apex * 0.1
        Dir = surf.Axis * 0.1
        tan = math.tan(surf.SemiAngle)
        X_dir = FreeCAD.Vector(1, 0, 0)
        Y_dir = FreeCAD.Vector(0, 1, 0)
        Z_dir = FreeCAD.Vector(0, 0, 1)
        if is_parallel(Dir, X_dir, tolerances.angle):
            sheet = 1
            if is_opposite(Dir, X_dir, tolerances.angle):
                sheet = -1
            if Apex.y == 0.0 and Apex.z == 0.0:
                mcnp_def = "{:<6d} KX  {:{x}} {:{t2}} {}".format(
                    id,
                    Apex.x,
                    tan**2,
                    sheet,
                    x=numeric_format.K_xyz,
                    t2=numeric_format.K_tan2,
                )
            else:
                mcnp_def = "{:<6d} K/X  {:{xyz}} {:{xyz}} {:{xyz}} {:{t2}} {}".format(
                    id,
                    Apex.x,
                    Apex.y,
                    Apex.z,
                    tan**2,
                    sheet,
                    xyz=numeric_format.K_xyz,
                    t2=numeric_format.K_tan2,
                )
        elif is_parallel(Dir, Y_dir, tolerances.angle):
            sheet = 1
            if is_opposite(Dir, Y_dir, tolerances.angle):
                sheet = -1
            if Apex.x == 0.0 and Apex.z == 0.0:
                mcnp_def = "{:<6d} KY  {:{y}} {:{t2}} {}".format(
                    id,
                    Apex.y,
                    tan**2,
                    sheet,
                    y=numeric_format.K_xyz,
                    t2=numeric_format.K_tan2,
                )
            else:
                mcnp_def = "{:<6d} K/Y  {:{xyz}} {:{xyz}} {:{xyz}} {:{t2}} {}".format(
                    id,
                    Apex.x,
                    Apex.y,
                    Apex.z,
                    tan**2,
                    sheet,
                    xyz=numeric_format.K_xyz,
                    t2=numeric_format.K_tan2,
                )
        elif is_parallel(Dir, Z_dir, tolerances.angle):
            sheet = 1
            if is_opposite(Dir, Z_dir, tolerances.angle):
                sheet = -1
            if Apex.x == 0.0 and Apex.y == 0.0:
                mcnp_def = "{:<6d} KZ  {:{z}} {:{t2}} {}".format(
                    id,
                    Apex.z,
                    tan**2,
                    sheet,
                    z=numeric_format.K_xyz,
                    t2=numeric_format.K_tan2,
                )
            else:
                mcnp_def = "{:<6d} K/Z  {:{xyz}} {:{xyz}} {:{xyz}} {:{t2}} {}".format(
                    id,
                    Apex.x,
                    Apex.y,
                    Apex.z,
                    tan**2,
                    sheet,
                    xyz=numeric_format.K_xyz,
                    t2=numeric_format.K_tan2,
                )
        else:
            Q = q_form.q_form_cone(Dir, Apex, tan)
            mcnp_def = """\
{:<6d} GQ  {v[0]:{aTof}} {v[1]:{aTof}} {v[2]:{aTof}}
          {v[3]:{aTof}} {v[4]:{aTof}} {v[5]:{aTof}}
          {v[6]:{gToi}} {v[7]:{gToi}} {v[8]:{gToi}}
          {v[9]:{j}} """.format(
                id,
                v=Q,
                aTof=numeric_format.GQ_1to6,
                gToi=numeric_format.GQ_7to9,
                j=numeric_format.GQ_10,
            )

    elif Type == "Sphere":
        # corresponding logic
        rad = surf.Radius * 0.1
        pnt = surf.Center * 0.1
        if pnt.isEqual(FreeCAD.Vector(0, 0, 0), tolerances.sph_distance):
            mcnp_def = "{:<6d} SO  {:{r}}".format(id, rad, r=numeric_format.S_r)
        else:
            mcnp_def = "{:<6d} S  {:{xyz}} {:{xyz}} {:{xyz}} {:{r}}".format(
                id,
                pnt.x,
                pnt.y,
                pnt.z,
                rad,
                xyz=numeric_format.S_xyz,
                r=numeric_format.S_r,
            )

    elif Type == "Torus":
        Dir = FreeCAD.Vector(surf.Axis)
        Dir.normalize()
        Pos = surf.Center * 0.1
        radMaj = surf.MajorRadius * 0.1
        radMin = surf.MinorRadius * 0.1
        if is_parallel(Dir, FreeCAD.Vector(1, 0, 0), tolerances.angle):
            mcnp_def = """\
{:<6d} TX  {:{xyz}} {:{xyz}} {:{xyz}}
           {:{r}} {:{r}} {:{r}}""".format(
                id,
                Pos.x,
                Pos.y,
                Pos.z,
                radMaj,
                radMin,
                radMin,
                xyz=numeric_format.T_xyz,
                r=numeric_format.T_r,
            )
        elif is_parallel(Dir, FreeCAD.Vector(0, 1, 0), tolerances.angle):
            mcnp_def = """\
{:<6d} TY  {:{xyz}} {:{xyz}} {:{xyz}}
           {:{r}} {:{r}} {:{r}}""".format(
                id,
                Pos.x,
                Pos.y,
                Pos.z,
                radMaj,
                radMin,
                radMin,
                xyz=numeric_format.T_xyz,
                r=numeric_format.T_r,
            )
        elif is_parallel(Dir, FreeCAD.Vector(0, 0, 1), tolerances.angle):
            mcnp_def = """\
{:<6d} TZ  {:{xyz}} {:{xyz}} {:{xyz}}
           {:{r}} {:{r}} {:{r}}""".format(
                id,
                Pos.x,
                Pos.y,
                Pos.z,
                radMaj,
                radMin,
                radMin,
                xyz=numeric_format.T_xyz,
                r=numeric_format.T_r,
            )

    return trim(mcnp_def, 80)


def open_mc_surface(Type, surf, tolerances, numeric_format, out_xml=True, quadricForm=False):
    if Type == "Plane":
        A = surf.Axis.x
        B = surf.Axis.y
        C = surf.Axis.z
        if surf.Axis.isEqual(FreeCAD.Vector(1, 0, 0), tolerances.pln_angle):
            D = surf.Position.x * 0.1
            if out_xml:
                omc_surf = "x-plane"
                coeffs = "{:{x}}".format(D, x=numeric_format.P_xyz)
            else:
                omc_surf = "XPlane"
                coeffs = f"x0={D}"

        elif surf.Axis.isEqual(FreeCAD.Vector(0, 1, 0), tolerances.pln_angle):
            D = surf.Position.y * 0.1
            if out_xml:
                omc_surf = "y-plane"
                coeffs = "{:{x}}".format(D, x=numeric_format.P_xyz)
            else:
                omc_surf = "YPlane"
                coeffs = f"y0={D}"

        elif surf.Axis.isEqual(FreeCAD.Vector(0, 0, 1), tolerances.pln_angle):
            D = surf.Position.z * 0.1
            if out_xml:
                omc_surf = "z-plane"
                coeffs = "{:{x}}".format(D, x=numeric_format.P_xyz)
            else:
                omc_surf = "ZPlane"
                coeffs = f"z0={D}"

        else:
            D = surf.Axis.dot(surf.Position) * 0.1
            if out_xml:
                omc_surf = "plane"
                coeffs = "{:{abc}} {:{abc}} {:{abc}} {:{d}}".format(A, B, C, D, abc=numeric_format.P_abc, d=numeric_format.P_d)
            else:
                omc_surf = "Plane"
                coeffs = f"a={A},b={B},c={C},d={D}"

    elif Type == "Cylinder":
        pos = surf.Center * 0.1
        Rad = surf.Radius * 0.1
        Dir = FreeCAD.Vector(surf.Axis)
        Dir.normalize()

        if is_parallel(Dir, FreeCAD.Vector(1, 0, 0), tolerances.angle):
            if out_xml:
                omc_surf = "x-cylinder"
                coeffs = "{:{xy}} {:{xy}} {:{r}}".format(pos.y, pos.z, Rad, xy=numeric_format.C_xyz, r=numeric_format.C_r)
            else:
                omc_surf = "XCylinder"
                coeffs = f"y0={pos.y},z0={pos.z},r={Rad}"

        elif is_parallel(Dir, FreeCAD.Vector(0, 1, 0), tolerances.angle):
            if out_xml:
                omc_surf = "y-cylinder"
                coeffs = "{:{xy}} {:{xy}} {:{r}}".format(pos.x, pos.z, Rad, xy=numeric_format.C_xyz, r=numeric_format.C_r)
            else:
                omc_surf = "YCylinder"
                coeffs = f"x0={pos.x},z0={pos.z},r={Rad}"

        elif is_parallel(Dir, FreeCAD.Vector(0, 0, 1), tolerances.angle):
            if out_xml:
                omc_surf = "z-cylinder"
                coeffs = "{:{xy}} {:{xy}} {:{r}}".format(pos.x, pos.y, Rad, xy=numeric_format.C_xyz, r=numeric_format.C_r)
            else:
                omc_surf = "ZCylinder"
                coeffs = f"x0={pos.x},y0={pos.y},r={Rad}"

        else:
            if out_xml:
                omc_surf = "quadric"
                Q = q_form.q_form_cyl(Dir, pos, Rad)
                coeffs = "{v[0]:{aTof}} {v[1]:{aTof}} {v[2]:{aTof}} {v[3]:{aTof}} {v[4]:{aTof}} {v[5]:{aTof}} {v[6]:{gToi}} {v[7]:{gToi}} {v[8]:{gToi}} {v[9]:{j}}".format(
                    v=Q,
                    aTof=numeric_format.GQ_1to6,
                    gToi=numeric_format.GQ_7to9,
                    j=numeric_format.GQ_10,
                )
            else:
                if quadricForm:
                    omc_surf = "Quadric"
                    Q = q_form.q_form_cyl(Dir, pos, Rad)
                    coeffs = "a={v[0]},b={v[1]},c={v[2]},d={v[3]},e={v[4]},f={v[5]},g={v[6]},h={v[7]},j={v[8]},k={v[9]}".format(
                        v=Q
                    )
                else:
                    omc_surf = "Cylinder"
                    coeffs = "x0={},y0={},z0={},r={},dx={},dy={},dz={}".format(pos.x, pos.y, pos.z, Rad, Dir.x, Dir.y, Dir.z)

    elif Type == "Cone":
        Apex = surf.Apex * 0.1
        Dir = FreeCAD.Vector(surf.Axis)
        Dir.normalize()
        tan = math.tan(surf.SemiAngle)
        tan2 = tan * tan

        X_dir = FreeCAD.Vector(1, 0, 0)
        Y_dir = FreeCAD.Vector(0, 1, 0)
        Z_dir = FreeCAD.Vector(0, 0, 1)

        if is_parallel(Dir, X_dir, tolerances.angle):
            if out_xml:
                omc_surf = "x-cone"
                coeffs = "{:{xyz}} {:{xyz}} {:{xyz}} {:{t2}}".format(
                    Apex.x,
                    Apex.y,
                    Apex.z,
                    tan2,
                    xyz=numeric_format.K_xyz,
                    t2=numeric_format.K_tan2,
                )
            else:
                omc_surf = "XCone"
                coeffs = f"x0={Apex.x},y0={Apex.y},z0={Apex.z},r2={tan2}"

        elif is_parallel(Dir, Y_dir, tolerances.angle):
            if out_xml:
                omc_surf = "y-cone"
                coeffs = "{:{xyz}} {:{xyz}} {:{xyz}} {:{t2}}".format(
                    Apex.x,
                    Apex.y,
                    Apex.z,
                    tan2,
                    xyz=numeric_format.K_xyz,
                    t2=numeric_format.K_tan2,
                )
            else:
                omc_surf = "YCone"
                coeffs = f"x0={Apex.x},y0={Apex.y},z0={Apex.z},r2={tan2}"

        elif is_parallel(Dir, Z_dir, tolerances.angle):
            if out_xml:
                omc_surf = "z-cone"
                coeffs = "{:{xyz}} {:{xyz}} {:{xyz}} {:{t2}}".format(
                    Apex.x,
                    Apex.y,
                    Apex.z,
                    tan2,
                    xyz=numeric_format.K_xyz,
                    t2=numeric_format.K_tan2,
                )
            else:
                omc_surf = "ZCone"
                coeffs = f"x0={Apex.x},y0={Apex.y},z0={Apex.z},r2={tan2}"

        else:
            if out_xml:
                omc_surf = "quadric"
                Q = q_form.q_form_cone(Dir, Apex, tan)
                coeffs = "{v[0]:{aTof}} {v[1]:{aTof}} {v[2]:{aTof}} {v[3]:{aTof}} {v[4]:{aTof}} {v[5]:{aTof}} {v[6]:{gToi}} {v[7]:{gToi}} {v[8]:{gToi}} {v[9]:{j}}".format(
                    v=Q,
                    aTof=numeric_format.GQ_1to6,
                    gToi=numeric_format.GQ_7to9,
                    j=numeric_format.GQ_10,
                )
            else:
                if quadricForm:
                    omc_surf = "Quadric"
                    Q = q_form.q_form_cone(Dir, Apex, tan)
                    coeffs = "a={v[0]},b={v[1]},c={v[2]},d={v[3]},e={v[4]},f={v[5]},g={v[6]},h={v[7]},j={v[8]},k={v[9]}".format(
                        v=Q
                    )
                else:
                    omc_surf = "Cone"
                    coeffs = "x0={},y0={},z0={},r2={},dx={},dy={},dz={}".format(
                        Apex.x, Apex.y, Apex.z, tan2, Dir.x, Dir.y, Dir.z
                    )

    elif Type == "Sphere":
        Center = surf.Center * 0.1
        Rad = surf.Radius * 0.1
        if out_xml:
            omc_surf = "sphere"
            coeffs = "{:{xyz}} {:{xyz}} {:{xyz}} {:{r}}".format(
                Center.x,
                Center.y,
                Center.z,
                Rad,
                xyz=numeric_format.S_xyz,
                r=numeric_format.S_r,
            )
        else:
            omc_surf = "Sphere"
            coeffs = f"x0={Center.x},y0={Center.y},z0={Center.z},r={Rad}"

    elif Type == "Torus":
        Center = surf.Center * 0.1
        minRad = surf.MinorRadius * 0.1
        majRad = surf.MajorRadius * 0.1
        Dir = FreeCAD.Vector(surf.Axis)
        Dir.normalize()
        if out_xml:
            coeffs = "{:{xyz}} {:{xyz}} {:{xyz}} {:{r}} {:{r}} {:{r}}".format(
                Center.x,
                Center.y,
                Center.z,
                majRad,
                minRad,
                minRad,
                xyz=numeric_format.T_xyz,
                r=numeric_format.T_r,
            )
        else:
            coeffs = "x0={},y0={},z0={},r={},r1={},r2={}".format(Center.x, Center.y, Center.z, majRad, minRad, minRad)

        if is_parallel(Dir, FreeCAD.Vector(1, 0, 0), tolerances.angle):
            omc_surf = "x-torus" if out_xml else "XTorus"
        elif is_parallel(Dir, FreeCAD.Vector(0, 1, 0), tolerances.angle):
            omc_surf = "y-torus" if out_xml else "YTorus"
        elif is_parallel(Dir, FreeCAD.Vector(0, 0, 1), tolerances.angle):
            omc_surf = "z-torus" if out_xml else "ZTorus"
        else:
            omc_surf = None

    if out_xml:
        coeffs = " ".join(coeffs.split())
    return omc_surf, coeffs


def serpent_surface(id, Type, surf, options, tolerance, numeric_format):
    serpent_def = ""

    if Type == "Plane":
        if surf.pointDef and options.prnt3PPlane:
            P1 = surf.Points[0]
            P2 = surf.Points[1]
            P3 = surf.Points[2]
            serpent_def = f"surf {id} plane {P1.x/10:{numeric_format.P_d}} {P1.y/10:{numeric_format.P_d}} {P1.z/10:{numeric_format.P_d}}\n"
            serpent_def += (
                f"      {P2.x/10:{numeric_format.P_d}} {P2.y/10:{numeric_format.P_d}} {P2.z/10:{numeric_format.P_d}}\n"
            )
            serpent_def += f"      {P3.x/10:{numeric_format.P_d}} {P3.y/10:{numeric_format.P_d}} {P3.z/10:{numeric_format.P_d}}"

        else:
            A = surf.Axis.x
            B = surf.Axis.y
            C = surf.Axis.z
            D = surf.Axis.dot(surf.Position)
            if surf.Axis.isEqual(FreeCAD.Vector(1, 0, 0), tolerance.pln_angle):
                serpent_def = f"surf {id} px {surf.Position.x/10:{numeric_format.P_xyz}}"
            elif surf.Axis.isEqual(FreeCAD.Vector(0, 1, 0), tolerance.pln_angle):
                serpent_def = f"surf {id} py {surf.Position.y/10:{numeric_format.P_xyz}}"
            elif surf.Axis.isEqual(FreeCAD.Vector(0, 0, 1), tolerance.pln_angle):
                serpent_def = f"surf {id} pz {surf.Position.z/10:{numeric_format.P_xyz}}"
            else:
                serpent_def = f"surf {id} plane {A:{numeric_format.P_d}} {B:{numeric_format.P_d}} {C:{numeric_format.P_d}} {D/10:{numeric_format.P_d}}"

    elif Type == "Cylinder":
        Dir = surf.Axis
        Dir.normalize()
        Pos = surf.Center * 0.1
        rad = surf.Radius * 0.1
        if is_parallel(Dir, FreeCAD.Vector(1, 0, 0), tolerance.angle):
            serpent_def = (
                f"surf {id} cylx {Pos.y:{numeric_format.C_xyz}} {Pos.z:{numeric_format.C_xyz}} {rad:{numeric_format.C_r}}"
            )
        elif is_parallel(Dir, FreeCAD.Vector(0, 1, 0), tolerance.angle):
            serpent_def = (
                f"surf {id} cyly {Pos.x:{numeric_format.C_xyz}} {Pos.z:{numeric_format.C_xyz}} {rad:{numeric_format.C_r}}"
            )
        elif is_parallel(Dir, FreeCAD.Vector(0, 0, 1), tolerance.angle):
            serpent_def = f"surf {id} cylz {rad:{numeric_format.C_r}}"
        else:
            # Is not still working fine
            Q = q_form.q_form_cyl(Dir, Pos, rad)
            serpent_def = """\
surf quadratic  {v[0]:{aTof}} {v[1]:{aTof}} {v[2]:{aTof}}
          {v[3]:{aTof}} {v[4]:{aTof}} {v[5]:{aTof}}
          {v[6]:{gToi}} {v[7]:{gToi}} {v[8]:{gToi}}
          {v[9]:{j}} """.format(
                id,
                v=Q,
                aTof=numeric_format.GQ_1to6,
                gToi=numeric_format.GQ_7to9,
                j=numeric_format.GQ_10,
            )

    elif Type == "Cone":
        Apex = surf.Apex * 0.1
        Dir = surf.Axis * 0.1
        tan = math.tan(surf.SemiAngle)
        X_dir = FreeCAD.Vector(1, 0, 0)
        Y_dir = FreeCAD.Vector(0, 1, 0)
        Z_dir = FreeCAD.Vector(0, 0, 1)

        # Need to check this
        # Serpent has no specific card for cone at origin, explicit origin only

        if is_parallel(Dir, X_dir, tolerance.angle):
            sheet = 1
            if is_opposite(Dir, X_dir, tolerance.angle):
                sheet = -1
            serpent_def = "surf ckx {:{xyz}} {:{xyz}} {:{xyz}} {:{t2}} {}".format(
                id,
                Apex.x,
                Apex.y,
                Apex.z,
                tan**2,
                sheet,
                xyz=numeric_format.K_xyz,
                t2=numeric_format.K_tan2,
            )
        elif is_parallel(Dir, Y_dir, tolerance.angle):
            sheet = 1
            if is_opposite(Dir, Y_dir, tolerance.angle):
                sheet = -1
            serpent_def = "surf cky {:{xyz}} {:{xyz}} {:{xyz}} {:{t2}} {}".format(
                id,
                Apex.x,
                Apex.y,
                Apex.z,
                tan**2,
                sheet,
                xyz=numeric_format.K_xyz,
                t2=numeric_format.K_tan2,
            )
        elif is_parallel(Dir, Z_dir, tolerance.angle):
            sheet = 1
            if is_opposite(Dir, Z_dir, tolerance.angle):
                sheet = -1
            serpent_def = "surf ckz {:{xyz}} {:{xyz}} {:{xyz}} {:{t2}} {}".format(
                id,
                Apex.x,
                Apex.y,
                Apex.z,
                tan**2,
                sheet,
                xyz=numeric_format.K_xyz,
                t2=numeric_format.K_tan2,
            )
        else:
            Q = q_form.q_form_cone(Dir, Apex, tan)

    elif Type == "Sphere":
        rad = surf.Radius * 0.1
        pnt = surf.Center * 0.1
        # Serpent has only explicit spheres at the origin
        serpent_def = f"surf {id} sph {pnt.x:{numeric_format.S_xyz}} {pnt.y:{numeric_format.S_xyz}} {pnt.z:{numeric_format.S_xyz}} {rad:{numeric_format.S_r}}"

    elif Type == "Torus":
        Dir = surf.Axis
        Dir.normalize()
        Pos = surf.Center * 0.1
        radMaj = surf.MajorRadius * 0.1
        radMin = surf.MinorRadius * 0.1
        if is_parallel(Dir, FreeCAD.Vector(1, 0, 0), tolerance.angle):
            serpent_def = (
                f"surf {id} torx {Pos.x:{numeric_format.T_xyz}} {Pos.y:{numeric_format.T_xyz}} {Pos.z:{numeric_format.T_xyz}}\n"
            )
            serpent_def += f"      {radMaj:{numeric_format.T_r}} {radMin:{numeric_format.T_r}} {radMin:{numeric_format.T_r}}"
        elif is_parallel(Dir, FreeCAD.Vector(0, 1, 0), tolerance.angle):
            serpent_def = (
                f"surf {id} tory {Pos.x:{numeric_format.T_xyz}} {Pos.y:{numeric_format.T_xyz}} {Pos.z:{numeric_format.T_xyz}}\n"
            )
            serpent_def += f"      {radMaj:{numeric_format.T_r}} {radMin:{numeric_format.T_r}} {radMin:{numeric_format.T_r}}"
        elif is_parallel(Dir, FreeCAD.Vector(0, 0, 1), tolerance.angle):
            serpent_def = (
                f"surf {id} torz {Pos.x:{numeric_format.T_xyz}} {Pos.y:{numeric_format.T_xyz}} {Pos.z:{numeric_format.T_xyz}}\n"
            )
            serpent_def += f"      {radMaj:{numeric_format.T_r}} {radMin:{numeric_format.T_r}} {radMin:{numeric_format.T_r}}"

    return serpent_def


def phits_surface(id, Type, surf, options, tolerance, numeric_format):
    phits_def = ""

    if Type == "Plane":
        if surf.pointDef and options.prnt3PPlane:
            P1 = surf.Points[0]
            P2 = surf.Points[1]
            P3 = surf.Points[2]
            phits_def = """{:<6d} P   {P1[0]:{d}} {P1[1]:{d}} {P1[2]:{d}} 
           {P2[0]:{d}} {P2[1]:{d}} {P2[2]:{d}}
           {P3[0]:{d}} {P3[1]:{d}} {P3[2]:{d}}""".format(
                id, P1=P1 / 10, P2=P2 / 10, P3=P3 / 10, d=numeric_format.P_d
            )
        else:
            A = surf.Axis.x
            B = surf.Axis.y
            C = surf.Axis.z
            if surf.Axis.isEqual(FreeCAD.Vector(1, 0, 0), tolerance.pln_angle):
                phits_def = "{:<6d} PX  {:{x}}".format(id, surf.Position.x / 10.0, x=numeric_format.P_xyz)
            elif surf.Axis.isEqual(FreeCAD.Vector(0, 1, 0), tolerance.pln_angle):
                phits_def = "{:<6d} PY  {:{y}}".format(id, surf.Position.y / 10.0, y=numeric_format.P_xyz)
            elif surf.Axis.isEqual(FreeCAD.Vector(0, 0, 1), tolerance.pln_angle):
                phits_def = "{:<6d} PZ  {:{z}}".format(id, surf.Position.z / 10.0, z=numeric_format.P_xyz)
            else:
                D = surf.Axis.dot(surf.Position)
                phits_def = "{:<6d} P   {:{abc}} {:{abc}} {:{abc}} {:{d}}".format(
                    id,
                    A,
                    B,
                    C,
                    D / 10.0,
                    abc=numeric_format.P_abc,
                    d=numeric_format.P_d,
                )

    elif Type == "Cylinder":
        Dir = FreeCAD.Vector(surf.Axis)
        Dir.normalize()
        Pos = surf.Center * 0.1
        rad = surf.Radius * 0.1
        if is_parallel(Dir, FreeCAD.Vector(1, 0, 0), tolerance.angle):
            if Pos.y == 0.0 and Pos.z == 0.0:
                phits_def = "{:<6d} CX  {:{r}}".format(id, rad, r=numeric_format.C_r)
            else:
                phits_def = "{:<6d} C/X  {:{yz}} {:{yz}} {:{r}}".format(
                    id, Pos.y, Pos.z, rad, yz=numeric_format.C_xyz, r=numeric_format.C_r
                )
        elif is_parallel(Dir, FreeCAD.Vector(0, 1, 0), tolerance.angle):
            if Pos.x == 0.0 and Pos.z == 0.0:
                phits_def = "{:<6d} CY  {:{r}}".format(id, rad, r=numeric_format.C_r)
            else:
                phits_def = "{:<6d} C/Y  {:{xz}} {:{xz}} {:{r}}".format(
                    id, Pos.x, Pos.z, rad, xz=numeric_format.C_xyz, r=numeric_format.C_r
                )
        elif is_parallel(Dir, FreeCAD.Vector(0, 0, 1), tolerance.angle):
            if Pos.y == 0.0 and Pos.x == 0.0:
                phits_def = "{:<6d} CZ  {:{r}}".format(id, rad, r=numeric_format.C_r)
            else:
                phits_def = "{:<6d} C/Z  {:{xy}} {:{xy}} {:{r}}".format(
                    id, Pos.x, Pos.y, rad, xy=numeric_format.C_xyz, r=numeric_format.C_r
                )
        else:
            # Is not still working fine
            Q = q_form.q_form_cyl(Dir, Pos, rad)
            phits_def = """\
{:<6d} GQ  {v[0]:{aTof}} {v[1]:{aTof}} {v[2]:{aTof}}
          {v[3]:{aTof}} {v[4]:{aTof}} {v[5]:{aTof}}
          {v[6]:{gToi}} {v[7]:{gToi}} {v[8]:{gToi}}
          {v[9]:{j}} """.format(
                id,
                v=Q,
                aTof=numeric_format.GQ_1to6,
                gToi=numeric_format.GQ_7to9,
                j=numeric_format.GQ_10,
            )

        # Si se quiere rcc en vez de Q form
        #    pnt = surf.Center.sub(surf.Axis.multiply(1.0e7)) # mas alla de 100 m
        #    dir = surf.Axis.multiply(1.0e8)
        #    Vx= pnt.x/10.0
        #    Vy= pnt.y/10.0
        #    Vz= pnt.z/10.0
        #    Hx= dir.x/10.0
        #    Hy= dir.y/10.0
        #    Hz= dir.z/10.0
        #    rad=surf.Radius/10.0
        #    mcnp_def='%i  RCC  %13.7E %13.7E %13.7E %13.7E\n       %13.7E %13.7E %13.7E' %(id,Vx,Vy,Vz,Hx,Hy,Hz,rad)

    elif Type == "Cone":
        Apex = surf.Apex * 0.1
        Dir = surf.Axis * 0.1
        tan = math.tan(surf.SemiAngle)
        X_dir = FreeCAD.Vector(1, 0, 0)
        Y_dir = FreeCAD.Vector(0, 1, 0)
        Z_dir = FreeCAD.Vector(0, 0, 1)
        if is_parallel(Dir, X_dir, tolerance.angle):
            sheet = 1
            if is_opposite(Dir, X_dir, tolerance.angle):
                sheet = -1
            if Apex.y == 0.0 and Apex.z == 0.0:
                phits_def = "{:<6d} KX  {:{x}} {:{t2}} {}".format(
                    id,
                    Apex.x,
                    tan**2,
                    sheet,
                    x=numeric_format.K_xyz,
                    t2=numeric_format.K_tan2,
                )
            else:
                phits_def = "{:<6d} K/X  {:{xyz}} {:{xyz}} {:{xyz}} {:{t2}} {}".format(
                    id,
                    Apex.x,
                    Apex.y,
                    Apex.z,
                    tan**2,
                    sheet,
                    xyz=numeric_format.K_xyz,
                    t2=numeric_format.K_tan2,
                )
        elif is_parallel(Dir, Y_dir, tolerance.angle):
            sheet = 1
            if is_opposite(Dir, Y_dir, tolerance.angle):
                sheet = -1
            if Apex.x == 0.0 and Apex.z == 0.0:
                phits_def = "{:<6d} KY  {:{y}} {:{t2}} {}".format(
                    id,
                    Apex.y,
                    tan**2,
                    sheet,
                    y=numeric_format.K_xyz,
                    t2=numeric_format.K_tan2,
                )
            else:
                phits_def = "{:<6d} K/Y  {:{xyz}} {:{xyz}} {:{xyz}} {:{t2}} {}".format(
                    id,
                    Apex.x,
                    Apex.y,
                    Apex.z,
                    tan**2,
                    sheet,
                    xyz=numeric_format.K_xyz,
                    t2=numeric_format.K_tan2,
                )
        elif is_parallel(Dir, Z_dir, tolerance.angle):
            sheet = 1
            if is_opposite(Dir, Z_dir, tolerance.angle):
                sheet = -1
            if Apex.x == 0.0 and Apex.y == 0.0:
                phits_def = "{:<6d} KZ  {:{z}} {:{t2}} {}".format(
                    id,
                    Apex.z,
                    tan**2,
                    sheet,
                    z=numeric_format.K_xyz,
                    t2=numeric_format.K_tan2,
                )
            else:
                phits_def = "{:<6d} K/Z  {:{xyz}} {:{xyz}} {:{xyz}} {:{t2}} {}".format(
                    id,
                    Apex.x,
                    Apex.y,
                    Apex.z,
                    tan**2,
                    sheet,
                    xyz=numeric_format.K_xyz,
                    t2=numeric_format.K_tan2,
                )
        else:
            Q = q_form.q_form_cone(Dir, Apex, tan)
            phits_def = """\
{:<6d} GQ  {v[0]:{aTof}} {v[1]:{aTof}} {v[2]:{aTof}}
          {v[3]:{aTof}} {v[4]:{aTof}} {v[5]:{aTof}}
          {v[6]:{gToi}} {v[7]:{gToi}} {v[8]:{gToi}}
          {v[9]:{j}} """.format(
                id,
                v=Q,
                aTof=numeric_format.GQ_1to6,
                gToi=numeric_format.GQ_7to9,
                j=numeric_format.GQ_10,
            )

    elif Type == "Sphere":
        # corresponding logic
        rad = surf.Radius * 0.1
        pnt = surf.Center * 0.1
        if pnt.isEqual(FreeCAD.Vector(0, 0, 0), tolerance.sph_distance):
            phits_def = "{:<6d} SO  {:{r}}".format(id, rad, r=numeric_format.S_r)
        else:
            phits_def = "{:<6d} S  {:{xyz}} {:{xyz}} {:{xyz}} {:{r}}".format(
                id,
                pnt.x,
                pnt.y,
                pnt.z,
                rad,
                xyz=numeric_format.S_xyz,
                r=numeric_format.S_r,
            )

    elif Type == "Torus":
        Dir = FreeCAD.Vector(surf.Axis)
        Dir.normalize()
        Pos = surf.Center * 0.1
        radMaj = surf.MajorRadius * 0.1
        radMin = surf.MinorRadius * 0.1
        if is_parallel(Dir, FreeCAD.Vector(1, 0, 0), tolerance.angle):
            phits_def = """\
{:<6d} TX  {:{xyz}} {:{xyz}} {:{xyz}}
           {:{r}} {:{r}} {:{r}}""".format(
                id,
                Pos.x,
                Pos.y,
                Pos.z,
                radMaj,
                radMin,
                radMin,
                xyz=numeric_format.T_xyz,
                r=numeric_format.T_r,
            )
        elif is_parallel(Dir, FreeCAD.Vector(0, 1, 0), tolerance.angle):
            phits_def = """\
{:<6d} TY  {:{xyz}} {:{xyz}} {:{xyz}}
           {:{r}} {:{r}} {:{r}}""".format(
                id,
                Pos.x,
                Pos.y,
                Pos.z,
                radMaj,
                radMin,
                radMin,
                xyz=numeric_format.T_xyz,
                r=numeric_format.T_r,
            )
        elif is_parallel(Dir, FreeCAD.Vector(0, 0, 1), tolerance.angle):
            phits_def = """\
{:<6d} TZ  {:{xyz}} {:{xyz}} {:{xyz}}
           {:{r}} {:{r}} {:{r}}""".format(
                id,
                Pos.x,
                Pos.y,
                Pos.z,
                radMaj,
                radMin,
                radMin,
                xyz=numeric_format.T_xyz,
                r=numeric_format.T_r,
            )

    return trim(phits_def, 80)


def trim(surfDef, lineLength=80):

    lines = surfDef.split("\n")
    if len(lines) == 1 and len(lines[0]) <= lineLength:
        return surfDef

    longLine = []
    for i, line in enumerate(lines):
        if len(line) > lineLength:
            longLine.append(i)

    if len(longLine) == 0:
        return surfDef

    newDef = ""
    for i, line in enumerate(lines):
        if i in longLine:
            newLine = cut_line(line, lineLength)
        else:
            newLine = line
        newDef += newLine + "\n"

    return newDef[:-1]


def cut_line(line, lineLength):
    tabNumber = 10
    while True:
        pos = line.rfind(" ")
        if pos <= lineLength:
            break

    line1 = line[0:pos]
    line2 = line[pos + 1 :]

    # second line is only spaces
    if len(line2.strip()) == 0:
        newLine = line1
    else:
        newLine = "{}\n{: <{n}}{}".format(line1, "", line2, n=tabNumber)

    return newLine



##############################
# Module to write MCNP input #
##############################
import logging
import math
from datetime import datetime
from pathlib import Path
from importlib.metadata import version

import FreeCAD

from ..utils.basic_functions_part1 import is_opposite, points_to_coeffs
from ..utils.functions import SurfacesDict
from .functions import CardLine, change_surf_sign, mcnp_surface, write_mcnp_cell_def

logger = logging.getLogger("general_logger")


# TODO rename as there are two classes with this name
class McnpInput:
    def __init__(
        self,
        Meta,
        Surfaces,
        options,
        tolerances,
        numeric_format,
        title,
        volSDEF,
        volCARD,
        UCARD,
        dummyMat,
        step_filename,
    ):
        self.Title = title
        self.VolSDEF = volSDEF
        self.VolCARD = volCARD
        self.U0CARD = UCARD
        self.dummyMat = dummyMat
        self.Cells = Meta
        self.options = options
        self.tolerances = tolerances
        self.numeric_format = numeric_format
        self.Options = {
            "Volume": self.VolCARD,
            "Particle": ("n", "p"),
            "Universe": self.U0CARD,
        }
        self.part = "P"

        self.step_filename = step_filename
        if isinstance(self.step_filename, (tuple, list)):
            self.step_filename = "; ".join(self.step_filename)

        self.get_surface_table()
        self.simplify_planes(Surfaces)

        self.Surfaces = self.sorted_surfaces(Surfaces)
        self.Materials = set()

        return

    def set_sdef(self, data):

        if data[0] is not None:
            sphereId = data[0][0]
            radius = data[0][1]
            wgt = math.pi * radius * radius * 1e-2
            sdef = "SDEF PAR={} NRM=-1 SUR={} WGT={:13.7e} DIR=d1\n".format(self.part, sphereId, wgt)
            SI1 = "SI1 0 1\n"
            SP1 = "SP1 -21 1\n"
            self.SDEF_sphere = (sdef, SI1, SP1)
        else:
            self.SDEF_sphere = None
        xmin, xmax, ymin, ymax, zmin, zmax = data[1]

        sdef = f"SDEF PAR={self.part} X=D1 Y=D2 Z=D3 \n"
        SI1 = f"SI1 {xmin * 0.1:13.7e} {xmax * 0.1:13.7e} \n"
        SI2 = f"SI2 {ymin * 0.1:13.7e} {ymax * 0.1:13.7e} \n"
        SI3 = f"SI3 {zmin * 0.1:13.7e} {zmax * 0.1:13.7e} \n"
        SP1 = "SP1 0  1 \n"
        SP2 = "SP2 0  1 \n"
        SP3 = "SP3 0  1 \n"
        self.SDEF_box = (sdef, SI1, SI2, SI3, SP1, SP2, SP3)

    def write_input(self, filename):
        Path(filename).parent.mkdir(parents=True, exist_ok=True)
        logger.info(f"write MCNP file {filename}")
        self.inpfile = open(filename, "w", encoding="utf-8")
        self.write_header()
        self.write_cell_block()
        self.inpfile.write(" \n")

        surfaceHeader = """\
C ##########################################################
C                  SURFACE DEFINITION
C ##########################################################
"""
        self.inpfile.write(surfaceHeader)
        self.write_surface_block()
        self.inpfile.write(" \n")

        self.write_source_block()

        self.inpfile.close()
        return

    def write_header(self):

        freeCAD_Version = "{V[0]:}.{V[1]:}.{V[2]:}".format(V=FreeCAD.Version())

        Header = f"""{self.Title}
C   ______ _______  _____      _     _ __   _ _______ ______  
C  |  ____ |______ |     | ___ |     | | \\  | |______ |     \\ 
C  |_____| |______ |_____|     |_____| |  \\_| |______ |_____/
C Version : {version('geouned')}
C FreeCAD Version : {freeCAD_Version} 
"""

        Information = f"""C
C *************************************************************
C Original Step file : {self.step_filename}
C
C Creation Date : {datetime.now()}
C Solid Cells   : {self.__solidCells__}
C Total Cells   : {self.__cells__}
C Surfaces      : {len(self.Surfaces)}
C Materials     : {len(self.__materials__)}
C
C **************************************************************
"""
        self.inpfile.write(Header)
        self.inpfile.write(Information)
        return

    def write_cell_block(self):

        for i, cell in enumerate(self.Cells):
            self.write_cells(cell)
        return

    def write_surface_block(self):

        for surf in self.Surfaces:
            self.write_surfaces(surf)

    def write_cells(self, cell):

        index = cell.label

        # if index is None objet not contain cell definition
        # but a comment to insert between cells
        if cell.__id__ is None:
            comment = self.comment_line(cell.Comments)
            self.inpfile.write(comment)
            return

        if cell.Material == 0:
            cellHeader = f"{index:<5d} {0:<5d}  "
        else:
            self.Materials.add(cell.Material)
            if abs(cell.Density) < 1e-2:
                cellHeader = f"{index:<5d} {cell.Material:<5d} {cell.Density:11.4e} "
            else:
                cellHeader = f"{index:<5d} {cell.Material:<5d} {cell.Density:11.7f} "

        mcnpcell = "{}{}\n{}{}".format(
            cellHeader,
            self.cell_format(cell.Definition, offset=len(cellHeader)),
            self.option_format(cell),
            self.comment_format(cell.Comments, cell.MatInfo),
        )
        self.inpfile.write(mcnpcell)
        return

    def write_surfaces(self, surface):
        """Write the surfaces in MCNP format"""

        MCNP_def = mcnp_surface(
            surface.Index,
            surface.Type,
            surface.Surf,
            self.options,
            self.tolerances,
            self.numeric_format,
        )
        if MCNP_def:
            MCNP_def += "\n"
            self.inpfile.write(MCNP_def)
        else:
            logger.info(f"Surface {surface.Type} cannot be written in MCNP input")
        return

    def write_source_block(self):

        if self.SDEF_sphere is None:
            return

        MODE = f"MODE {self.part}\nVOID \nNPS 1e6\n"
        if self.dummyMat:
            mat = list(self.Materials)
            mat.sort()
            MATCARD = ""
            for m in mat:
                MATCARD += f"M{m:<6d} 1001 1\n"
            Block = MATCARD + "C \n" + MODE
        else:
            Block = MODE

        if self.VolSDEF:
            Block += "PRDMP 2J -1\n"
            for line in self.SDEF_box:
                Block += "C " + line
            for line in self.SDEF_sphere:
                Block += line

            celList, volList = self.get_solid_cell_volume()

            F4Tally = CardLine(f"F4:{self.part} ")
            F4Tally.extend(celList)
            SD4 = CardLine("SD4  ", fmt="13.7e")
            SD4.extend(volList)

            Block += F4Tally.str + "\n"
            if not self.VolCARD:
                Block += SD4.str
            else:
                Block += "C Cell volume normalization is set in cell cards VOL\n"
        else:
            for line in self.SDEF_sphere:
                Block += "C " + line
            for line in self.SDEF_box:
                Block += line

        self.inpfile.write(Block)

    def cell_format(self, Definition, offset=11):
        return write_mcnp_cell_def(Definition, tabspace=11, offset=offset)

    def option_format(self, cell):

        option = ""
        if self.Options["Volume"]:
            if not cell.Void:
                option = f"{'':11s}Vol={cell.Volume * 0.001:e}\n"
            else:
                option = f"{'':11s}Vol=1.0\n"

        option += f"{'':11s}"
        for p in self.Options["Particle"]:
            if cell.MatInfo == "Graveyard":
                option += f"imp:{p}=0     "
            else:
                option += f"imp:{p}=1.0   "

        if self.Options["Universe"] is not None:
            option += f"U={self.Options['Universe']}  "
        option += "\n"

        return option

    def comment_format(self, cComment, mComment=None):

        comment = ""
        if mComment:
            mComment = mComment.split("\n")
            for c in mComment:
                if c:
                    comment += f"{'':11s}${c}\n"

        if cComment.strip() != "":
            cComment = cComment.strip().split("\n")
            for c in cComment:
                if c:
                    comment += f"{'':11s}${c}\n"
        return comment

    def comment_line(self, lineComment):
        lineComment = lineComment.strip().split("\n")
        comment = ""
        if lineComment:
            comment = "C \n"
            for c in lineComment:
                if c:
                    comment += f"C {c}\n"
            comment += "C \n"
        return comment

    def get_surface_table(self):
        self.surfaceTable = {}
        self.__solidCells__ = 0
        self.__cells__ = 0
        self.__materials__ = set()

        for i, CellObj in enumerate(self.Cells):
            if CellObj.__id__ is None:
                continue
            self.__cells__ += 1
            if CellObj.Material != 0:
                self.__materials__.add(CellObj.Material)

            surf = CellObj.Definition.get_surfaces_numbers()
            if not CellObj.Void:
                self.__solidCells__ += 1
            for index in surf:
                if index in self.surfaceTable.keys():
                    self.surfaceTable[index].add(i)
                else:
                    self.surfaceTable[index] = {i}
        return

    def simplify_planes(self, Surfaces):

        for p in Surfaces["PX"]:
            if p.Surf.Axis[0] < 0:
                p.Surf.Axis = FreeCAD.Vector(1, 0, 0)
                self.change_surf_sign(p)

        for p in Surfaces["PY"]:
            if p.Surf.Axis[1] < 0:
                p.Surf.Axis = FreeCAD.Vector(0, 1, 0)
                self.change_surf_sign(p)

        for p in Surfaces["PZ"]:
            if p.Surf.Axis[2] < 0:
                p.Surf.Axis = FreeCAD.Vector(0, 0, 1)
                self.change_surf_sign(p)

        if self.options.prnt3PPlane:
            for p in Surfaces["P"]:
                if p.Surf.pointDef:
                    axis, d = points_to_coeffs(p.Surf.Points)
                    if is_opposite(axis, p.Surf.Axis):
                        self.change_surf_sign(p)

        return

    def sorted_surfaces(self, Surfaces):
        temp = SurfacesDict(Surfaces)
        surfList = []
        for ind in range(Surfaces.IndexOffset, Surfaces.surfaceNumber + Surfaces.IndexOffset):
            s = temp.get_surface(ind + 1)
            if s is not None:
                surfList.append(s)
                temp.del_surface(ind + 1)
        return surfList

    def change_surf_sign(self, p):

        if p.Index not in self.surfaceTable.keys():
            logger.info(f"{p.Type} Surface {p.Index} not used in cell definition) {p.Surf.Axis} {p.Surf.Position}")
            return

        for ic in self.surfaceTable[p.Index]:
            surf = self.Cells[ic].Definition.get_surfaces_numbers()
            for s in surf:
                if s == p.Index:
                    change_surf_sign(s, self.Cells[ic].Definition)

    def get_solid_cell_volume(self):

        solidList = []
        volumeList = []
        for m in self.Cells:
            if m.CellType == "solid" and m.__id__ is not None:
                solidList.append(m.label)
                volumeList.append(m.Volume * 1e-3)

        return solidList, volumeList



##############################
# Module to write MCNP input #
##############################

import logging
from pathlib import Path

import FreeCAD

from ..code_version import *
from ..utils.functions import SurfacesDict
from .functions import change_surf_sign, open_mc_surface, write_openmc_region

logger = logging.getLogger("general_logger")


class OpenmcInput:
    def __init__(self, Meta, Surfaces, options, tolerances, numeric_format):

        self.Cells = Meta
        self.options = options
        self.tolerances = tolerances
        self.numeric_format = numeric_format

        self.get_surface_table()
        self.simplify_planes(Surfaces)

        self.Surfaces = self.sorted_surfaces(Surfaces)
        self.Materials = set()

    def write_xml(self, filename):
        logger.info(f"write OpenMC xml file {filename}")
        Path(filename).parent.mkdir(parents=True, exist_ok=True)
        with open(file=filename, mode="w", encoding="utf-8") as self.inpfile:
            self.write_xml_header()
            self.inpfile.write("<geometry>\n")
            self.write_xml_cell_block()
            self.inpfile.write(" \n")
            self.write_xml_surface_block()
            self.inpfile.write("</geometry>\n")
        return

    def write_xml_header(self):
        Header = "<?xml version='1.0' encoding='UTF-8'?>\n"
        self.inpfile.write(Header)
        return

    def write_xml_cell_block(self):
        for _, cell in enumerate(self.Cells):
            if cell.MatInfo == "Graveyard":
                continue
            self.write_xml_cells(cell)
        return

    def write_xml_surface_block(self):
        for surf in self.Surfaces[:-1]:
            self.write_xml_surfaces(surf)
        self.write_xml_surfaces(self.Surfaces[-1], boundary=True)

    def write_xml_cells(self, cell):
        """Write the cell in xml OpenMC format"""
        index = cell.label
        cellName = ". ".join(cell.Comments.splitlines())
        if cell.__id__ is None:
            return

        if cell.Material == 0:
            matName = "void"
        else:
            matName = f"{cell.Material}"

        OMCcell = '  <cell id="{}" material="{}" name="{}" region="{}" universe="1"/>\n'.format(
            index,
            matName,
            cellName,
            write_openmc_region(cell.Definition, self.options, "XML"),
        )
        self.inpfile.write(OMCcell)
        return

    def write_xml_surfaces(self, surface, boundary=False):
        """Write the surfaces in xml OpenMC format"""

        surfType, coeffs = open_mc_surface(surface.Type, surface.Surf, self.tolerances, self.numeric_format)

        if not boundary:
            OMCsurf = '  <surface id="{}" type="{}" coeffs="{}" />\n'.format(surface.Index, surfType, coeffs)
        else:
            OMCsurf = '  <surface id="{}" type="{}" coeffs="{}" boundary="vacuum" />\n'.format(surface.Index, surfType, coeffs)

        self.inpfile.write(OMCsurf)
        return

    def write_py(self, filename):
        logger.info(f"write OpenMC python script {filename}")

        # get all the materials present in the model
        for cell in self.Cells:
            if cell.Material != 0:
                self.Materials.add(cell.Material)

        Path(filename).parent.mkdir(parents=True, exist_ok=True)
        with open(file=filename, mode="w", encoding="utf-8") as self.inpfile:
            self.write_py_header()

            if len(self.Materials) > 0:
                self.inpfile.write("# Materials setup\n")
                self.write_py_materials()
            self.inpfile.write("\n")

            self.inpfile.write("# Surface setup\n")
            self.write_py_surface_block()
            self.inpfile.write("\n")

            self.inpfile.write("# Cell definition \n")
            self.write_py_cell_block()
        return

    def write_py_header(self):

        Header = """\
# openMC geometry script generated by GEOUNED
import openmc

###############################################################################
# Define problem geometry
###############################################################################

"""
        self.inpfile.write(Header)
        return

    def write_py_materials(self):
        matList = tuple(sorted(self.Materials))
        strMat = []
        for m in matList:
            material = f"M{m} = openmc.Material(name='M{m}')\n"
            self.inpfile.write(material)
            strMat.append(f"M{m}")

        collect = f"materials = openmc.Materials([{', '.join(strMat)}])\n"
        self.inpfile.write(collect)
        self.inpfile.write("materials.export_to_xml()\n")

    def write_py_surface_block(self):

        for surf in self.Surfaces[:-1]:
            self.write_py_surfaces(surf)
        self.write_py_surfaces(self.Surfaces[-1], boundary=True)

    def write_py_surfaces(self, surface, boundary=False):
        """Write the surfaces in python OpenMC format"""

        surfType, coeffs = open_mc_surface(
            surface.Type,
            surface.Surf,
            self.tolerances,
            self.numeric_format,
            out_xml=False,
            quadricForm=self.options.quadricPY,
        )

        if not boundary:
            OMCsurf = f"S{surface.Index} = openmc.{surfType}({coeffs})\n"
        else:
            OMCsurf = 'S{} = openmc.{}({}, boundary_type="vacuum")\n'.format(surface.Index, surfType, coeffs)

        self.inpfile.write(OMCsurf)
        return

    def write_py_cell_block(self):

        cellNames = []
        for i, cell in enumerate(self.Cells):
            if cell.MatInfo == "Graveyard":
                continue
            self.write_py_cells(cell)
            if cell.__id__ is None:
                continue
            cellNames.append(f"C{cell.label}")

        geometry = "\ngeometry = openmc.Geometry([{}])\ngeometry.export_to_xml()\n".format(", ".join(cellNames))

        self.inpfile.write(geometry)
        return

    def write_py_cells(self, cell):
        """Write the cell in python OpenMC format"""
        index = cell.label
        cellName = ". ".join(cell.Comments.splitlines())
        if cell.__id__ is None:
            return

        if cell.Material == 0:
            OMCcell = 'C{} = openmc.Cell(name="{}", region={})\n'.format(
                index,
                cellName,
                write_openmc_region(cell.Definition, self.options, "PY"),
            )
        else:
            matName = f"M{cell.Material}"
            OMCcell = 'C{} = openmc.Cell(name="{}", fill={}, region={})\n'.format(
                index,
                cellName,
                matName,
                write_openmc_region(cell.Definition, self.options, "PY"),
            )
        self.inpfile.write(OMCcell)
        return

    def get_surface_table(self):
        self.surfaceTable = {}
        self.__solidCells__ = 0
        self.__cells__ = 0
        self.__materials__ = set()

        for i, CellObj in enumerate(self.Cells):
            if CellObj.__id__ is None:
                continue
            self.__cells__ += 1
            if CellObj.Material != 0:
                self.__materials__.add(CellObj.Material)

            surf = CellObj.Definition.get_surfaces_numbers()
            if not CellObj.Void:
                self.__solidCells__ += 1
            for index in surf:
                if index in self.surfaceTable.keys():
                    self.surfaceTable[index].add(i)
                else:
                    self.surfaceTable[index] = {i}
        return

    def simplify_planes(self, Surfaces):

        for p in Surfaces["PX"]:
            if p.Surf.Axis[0] < 0:
                p.Surf.Axis = FreeCAD.Vector(1, 0, 0)
                self.change_surf_sign(p)

        for p in Surfaces["PY"]:
            if p.Surf.Axis[1] < 0:
                p.Surf.Axis = FreeCAD.Vector(0, 1, 0)
                self.change_surf_sign(p)

        for p in Surfaces["PZ"]:
            if p.Surf.Axis[2] < 0:
                p.Surf.Axis = FreeCAD.Vector(0, 0, 1)
                self.change_surf_sign(p)
        return

    def sorted_surfaces(self, Surfaces):
        temp = SurfacesDict(Surfaces)
        surfList = []
        for ind in range(Surfaces.IndexOffset, Surfaces.surfaceNumber + Surfaces.IndexOffset):
            s = temp.get_surface(ind + 1)
            if s is not None:
                surfList.append(s)
                temp.del_surface(ind + 1)
        return surfList

    def change_surf_sign(self, p):

        if p.Index not in self.surfaceTable.keys():
            logger.info(f"{p.Type} Surface {p.Index} not used in cell definition {p.Surf.Axis} {p.Surf.Position}")
            return

        for ic in self.surfaceTable[p.Index]:
            surf = self.Cells[ic].Definition.get_surfaces_numbers()
            for s in surf:
                if s == p.Index:
                    change_surf_sign(s, self.Cells[ic].Definition)



from importlib.metadata import version

# this try except attempts to import freecad (lowercase) which is the conda
# package name for FreeCAD (mixed case) upon import the conda package appends
# the sys path for Conda installed FreeCAD, consequently FreeCAD can then be
# found by subsequent import statements through out the code base
try:
    import freecad
except ImportError:
    pass

from .GEOReverse import *
from .GEOUNED import *
from .GEOUNED.utils.log_utils import setup_logger

setup_logger("general_logger", "geouned_general_log.log")
setup_logger("fuzzy_logger", "geouned_fuzzy_log.log")
setup_logger("solids_logger", "geouned_solids_log.log")


__version__ = version("geouned")

__all__ = ["__version__"]



#
#  definition of GEOUNED objects to release memory
#
#  GEOUNED SurfacesGU, SolidsGU, PlaneGU, etc.. objects are created because FreeCAD create a new object
#  each time an attribute of FreeCAD object is called. This leads to code crash with memory failure
#  when attribues are call large amount of times. Like it is in this code.

import logging
import math

import Part

from .basic_functions_part1 import is_same_value
from .basic_functions_part2 import is_same_torus

logger = logging.getLogger("general_logger")


# SURFACES
class SurfacesGu(object):
    """GEOUNED surface class"""

    def __init__(self, face):
        self.face = face
        self.Surface = self.face.Surface
        self.type = str(self.Surface)

    def __str__(self):
        """str(Surface) is done for the classification of the surface.
        Surface_GU saves this information in self.type"""
        return self.type


class PlaneGu(SurfacesGu):
    """GEOUNED Plane Class"""

    def __init__(self, face, plane3Pts=False):
        SurfacesGu.__init__(self, face)
        self.Axis = face.Surface.Axis
        self.Position = face.Surface.Position
        self.pointDef = plane3Pts
        if plane3Pts:
            self.Points = tuple(v.Point for v in face.Vertexes)
            d1 = self.Points[0] - self.Points[1]
            d2 = self.Points[0] - self.Points[2]
            d3 = self.Points[1] - self.Points[2]
            self.dim1 = max(d1.Length, d2.Length, d3.Length)
            self.dim2 = min(d1.Length, d2.Length, d3.Length)
        else:
            self.dim1 = face.ParameterRange[1] - face.ParameterRange[0]
            self.dim2 = face.ParameterRange[3] - face.ParameterRange[2]

    def isSameSurface(self, surface):
        if type(surface) is not PlaneGu:
            return False
        if abs(self.Axis.dot(surface.Axis)) < 0.99999:
            return False
        if abs(self.Axis.dot(self.Position) - surface.Axis.dot(surface.Position)) > 1e-5:
            return False
        return True


class CylinderGu(SurfacesGu):
    """GEOUNED Cylinder Class"""

    def __init__(self, face):
        SurfacesGu.__init__(self, face)
        self.Axis = face.Surface.Axis
        self.Radius = face.Surface.Radius
        self.Center = face.Surface.Center
        self.dimL = face.ParameterRange[3] - face.ParameterRange[2]

    def isSameSurface(self, surface):
        if type(surface) is not CylinderGu:
            return False
        if abs(self.Radius - surface.Radius) > 1e-5:
            return False
        d = self.Center - surface.Center
        if d.Length > 1e-5:
            return False
        if abs(self.Axis.dot(surface.Axis)) < 0.99999:
            return False
        return True


class ConeGu(SurfacesGu):
    """GEOUNED Cone Class"""

    def __init__(self, face):
        SurfacesGu.__init__(self, face)
        self.Axis = face.Surface.Axis
        self.Apex = face.Surface.Apex
        self.SemiAngle = face.Surface.SemiAngle
        self.dimL = face.ParameterRange[3] - face.ParameterRange[2]
        self.dimR = face.Surface.Radius
        self.Radius = face.Surface.Radius

    def isSameSurface(self, surface):
        if type(surface) is not ConeGu:
            return False
        if abs(self.SemiAngle - surface.SemiAngle) > 1e-5:
            return False
        d = self.Apex - surface.Apex
        if d.Length > 1e-5:
            return False
        if abs(self.Axis.dot(surface.Axis)) < 0.99999:
            return False
        return True


class SphereGu(SurfacesGu):
    """GEOUNED Sphere Class"""

    def __init__(self, face):
        SurfacesGu.__init__(self, face)
        self.type = self.type[0:6]
        self.Center = face.Surface.Center
        self.Radius = face.Surface.Radius

    def isSameSurface(self, surface):
        if type(surface) is not SphereGu:
            return False
        if abs(self.Radius - surface.Radius) > 1e-5:
            return False
        d = self.Center - surface.Center
        if d.Length > 1e-5:
            return False
        return True


class TorusGu(SurfacesGu):
    """GEOUNED Torus Class"""

    def __init__(self, face):
        SurfacesGu.__init__(self, face)
        self.Center = face.Surface.Center
        self.Axis = face.Surface.Axis
        self.MajorRadius = face.Surface.MajorRadius
        self.MinorRadius = face.Surface.MinorRadius

    def isSameSurface(self, surface):
        if type(surface) is not TorusGu:
            return False
        if abs(self.MajorRadius - surface.MajorRadius) > 1e-5:
            return False
        if abs(self.MinorRadius - surface.MinorRadius) > 1e-5:
            return False
        d = self.Center - surface.Center
        if d.Length > 1e-5:
            return False
        if abs(self.Axis.dot(surface.Axis)) < 0.99999:
            return False
        return True


class SolidGu:
    """GEOUNED Solid Class"""

    def __init__(self, solid, tolerances, plane3Pts=False):
        self.solid = solid
        faces = define_list_face_gu(solid.Faces, plane3Pts)
        self.Faces = faces
        self.tolerances = tolerances
        self.Solids = solid.Solids
        self.BoundBox = solid.BoundBox
        self.Edges = solid.Edges
        self.TorusVParams = {}
        self.TorusUParams = {}

        toroidIndex = []
        for i, face in enumerate(self.Faces):
            if type(face.Surface) is TorusGu:
                toroidIndex.append(i)

        if len(toroidIndex) != 0:
            tFaces = self.same_torus_surf(toroidIndex)
            for i, tSet in enumerate(tFaces):
                URange = self.merge_periodic_uv("U", tSet)
                VRange = self.merge_periodic_uv("V", tSet)
                for t in tSet:
                    self.TorusVParams[t] = (i, VRange)
                    self.TorusUParams[t] = (i, URange)

    def same_torus_surf(self, torusList):
        """group as a single face all the neighbour faces of the same torus"""
        sameTorusFace = []
        temp = torusList[:]
        while len(temp) > 0:
            i = temp[0]
            current = [i]
            for j in temp[1:]:
                if is_same_torus(
                    self.Faces[i].Surface,
                    self.Faces[j].Surface,
                    dtol=self.tolerances.tor_distance,
                    atol=self.tolerances.tor_angle,
                    rel_tol=self.tolerances.relativeTol,
                ):
                    current.append(j)
            for c in current:
                temp.remove(c)
            sameTorusFace.append(current)

        return self.separate_surfaces(sameTorusFace)

    def separate_surfaces(self, faceList):
        """group all faces in faceList forming a continuous surface"""
        sameSurfaces = []
        for tset in faceList:
            temp = tset[:]
            while len(temp) > 0:
                i = 0
                current = [temp[0]]
                removeList = [temp[0]]
                while len(temp) > 0 and i < len(current):
                    for tindex in temp:
                        if self.Faces[current[i]].distToShape(self.Faces[tindex])[0] < self.tolerances.distance:
                            if tindex not in current:
                                current.append(tindex)
                                removeList.append(tindex)
                    i += 1
                    for c in removeList:
                        temp.remove(c)
                    removeList = []

                sameSurfaces.append(current)
        return sameSurfaces

    # TODO check if this function is used as it appears to be nut used in the code
    def merge_no_periodic_uv(self, parameter, faceList):
        if parameter == "U":
            i1 = 0
            i2 = 2
        elif parameter == "V":
            i1 = 2
            i2 = 4

        v_min, v_max = self.Faces[faceList[0]].ParameterRange[i1:i2]
        for face in faceList[1:]:
            V0, V1 = self.Faces[face].ParameterRange[i1:i2]
            v_min = min(v_min, V0)
            v_max = max(v_max, V1)
        mergedParams = (False, (v_min, v_max))

        return mergedParams

    def merge_periodic_uv(self, parameter, faceList):
        two_pi = 2.0 * math.pi
        if parameter == "U":
            i1 = 0
            i2 = 2
        elif parameter == "V":
            i1 = 2
            i2 = 4

        params = []
        arcLength = 0.0
        for face in faceList:
            V0, V1 = self.Faces[face].ParameterRange[i1:i2]
            arcLength += V1 - V0
            params.append((V0, V1))

        params.sort()
        V0 = params[0][0]
        V1 = params[-1][1]
        if arcLength >= two_pi * (1.0 - self.tolerances.relativePrecision):
            mergedParams = (True, (V0, V0 + two_pi))
        else:
            if is_same_value(V0, 0.0, self.tolerances.relativePrecision) and is_same_value(
                V1, two_pi, self.tolerances.relativePrecision
            ):
                for i in range(len(params) - 1):
                    if not is_same_value(
                        params[i][1],
                        params[i + 1][0],
                        self.tolerances.relativePrecision,
                    ):
                        break
                v_min = params[i + 1][0] - two_pi
                v_max = params[i][1]
            else:
                v_min = params[0][0]
                v_max = params[-1][1]
            mergedParams = (False, (v_min, v_max))

        return mergedParams


# FACES
class FaceGu(object):
    """GEOUNED Face Class"""

    def __init__(self, face, Plane3Pts=False):
        # GEOUNED based atributes
        self.__face__ = face
        self.Surface = define_surface(face, Plane3Pts)  # Define the appropiate GU Surface of the face

        # FreeCAD based Atributes
        self.Area = face.Area
        self.CenterOfMass = face.CenterOfMass
        self.ParameterRange = face.ParameterRange
        self.Orientation = face.Orientation
        self.OuterWire = face.OuterWire
        self.Edges = face.Edges
        self.Vertexes = face.Vertexes
        return

    def tessellate(self, val, reset=False):
        res = self.__face__.tessellate(val, reset)
        return res

    def getUVNodes(self):
        return self.__face__.getUVNodes()

    def isEqual(self, face):
        return self.__face__.isEqual(face.__face__)

    def isSame(self, face):
        return self.__face__.isSame(face.__face__)

    def valueAt(self, u, v):
        return self.__face__.valueAt(u, v)

    def distToShape(self, shape):
        shape1 = self.__face__
        if type(shape) is Part.Shape:
            shape2 = shape
        else:
            shape2 = shape.__face__

        if shape1 is shape2:
            return (0,)
        else:
            try:
                dist2Shape = shape1.distToShape(shape2)
            except:
                dist2Shape = shape2.distToShape(shape1)
            return dist2Shape


# Aux functions
def define_list_face_gu(face_list, plane3Pts=False):
    """Return the list of the  corresponding Face_GU  object of a FaceList"""
    return tuple(FaceGu(face, plane3Pts) for face in face_list)


def define_surface(face, plane3Pts):
    kind_surf = type(face.Surface)
    if kind_surf is Part.Plane:
        Surf_GU = PlaneGu(face, plane3Pts)
    elif kind_surf is Part.Cylinder:
        Surf_GU = CylinderGu(face)
    elif kind_surf is Part.Cone:
        Surf_GU = ConeGu(face)
    elif kind_surf is Part.Sphere:
        Surf_GU = SphereGu(face)
    elif kind_surf is Part.Toroid:
        Surf_GU = TorusGu(face)
    else:
        logger.info(f"bad Surface type {kind_surf}")
        Surf_GU = None
    return Surf_GU



import Part

from .options import Options
from .splitFunction import SplitBase, SplitSolid, joinBase
from .Utils.booleanFunction import BoolSequence


def getPart(slist):
    sol = []
    for s in slist:
        if type(s) is list:
            sol.extend(getPart(s))
        else:
            sol.append(s)
    return sol


def BuildSolid(cell, boundBox, mode="oneByOne", simplify=False):

    cutCell = cell.makeBox(boundBox)
    # cell.definition = BoolSequence(cell.definition.str)
    cell.cleanUndefined()

    celParts = BuildDepth(cell, SplitBase(cutCell), mode, True, simplify)

    celParts = getPart(celParts)
    # print('celparts',len(celParts))
    shapeParts = []
    for i, s in enumerate(celParts):
        shapeParts.append(s.base)
        # s.base.exportStep('solid{}.stp'.format(i))

    #   tt = FuseSolid(shapeParts)
    #   tt = tt.removeSplitter()
    # tt=Part.makeCompound(shapeParts)
    # tt.exportStep('cell{}.stp'.format(cell.name))
    return shapeParts
    # return FuseSolid(shapeParts)


def BuildDepth(cell, cutShape, mode, baseBox, simplify=False, loop=0):

    loop += 1
    seq = cell.definition
    if seq.level == 0:
        if baseBox:
            cutShape, cut = BuildSolidParts(cell, cutShape, mode)
        else:
            cutShape, cut = BuildSolidParts(cell, cutShape, "solids")
        return cutShape

    if type(cutShape) is not list:
        cutShape = [cutShape]
    newCutShape = []
    for i, CS in enumerate(cutShape):
        cbaseBox = baseBox
        # CS.base.exportStep('CS_{}_{}.stp'.format(i,str(cell.definition)))
        # CTable =build_c_table_from_solids(cell.makeBox(CS.base.BoundBox),cell.surfaces,option='full')
        # cell.definition.simplify(CTable)
        cell.definition.group_single()

        if type(cell.definition.elements) is not bool:
            if cell.definition.level == 0:
                tmp = BoolSequence(operator=cell.definition.operator)
                tmp.append(cell.definition)
                cell.definition = tmp

            if seq.operator == "AND":
                part = CS
                for e in cell.definition.elements:
                    part = BuildDepth(cell.getSubCell(e), part, mode, cbaseBox, simplify, loop=loop)
                    cbaseBox = False
                newCutShape.extend(part)
            else:
                cellParts = []
                for e in cell.definition.elements:
                    sub = cell.getSubCell(e)
                    part = BuildDepth(sub, CS, mode, baseBox, simplify, loop=loop)
                    cellParts.extend(part)

                JB = joinBase(cellParts)
                if JB.base is not None:
                    newCutShape.append(JB)

        elif cell.definition.elements:
            newCutShape.append(CS)

    cutShape = newCutShape
    return cutShape


def BuildSolidParts(cell, base, mode):

    # part if several base in input
    if type(base) is list or type(base) is tuple:
        fullPart = []
        cutPart = []
        for b in base:
            fullList, cutList = BuildSolidParts(cell, b, mode)
            fullPart.extend(fullList)
            cutPart.extend(cutList)
        return fullPart, cutPart

    boundBox = base.base.BoundBox
    surfaces = tuple(cell.surfaces.values())
    # print('\nbuild Parts :',mode)
    # print(cell.definition)
    # print(boundBox)

    if mode == "solids":

        # TODO consider making this buildShape call conditional
        # if cell.definition.operator == "OR" and False:
        #     Def = cell.definition
        #     cell.definition = cell.definition.get_complementary()
        #     cell.buildShape(boundBox, force=False, simplify=False)
        #     cell.definition = Def
        # else:
        #     cell.buildShape(boundBox, force=True, simplify=False, fuse=True)
        cell.buildShape(boundBox, force=True, simplify=False, fuse=True)

        # print('export')
        # base.base.exportStep('base.stp')
        # name=''.join(str(cell.definition))
        # cell.shape.exportStep('sol{}.stp'.format(name))
    else:
        cell.buildSurfaceShape(boundBox)

    if not surfaces:
        print("not cutting surfaces")
        return tuple(base.base), tuple()
    if mode == "solids":
        full, cut = SplitSolid(base, surfaces, cell, solidTool=True, tolerance=Options.splitTolerance)
    elif mode == "allSurfaces":
        full, cut = SplitSolid(base, surfaces, cell, tolerance=Options.splitTolerance)

    elif mode == "planeFirst":
        planes = []
        others = []
        for s in surfaces:
            # s.buildShape( boundBox)
            # s.shape.exportStep('Tool_{}_{}.stp'.format(s.type,s.id))
            if s.type == "plane":
                planes.append(s)
            else:
                others.append(s)

        if planes:

            full, cut = SplitSolid(base, planes, cell, tolerance=Options.splitTolerance)
            # for i,s in enumerate(full):
            #    s.exportStep('fullplane_{}.stp'.format(i))
            # for i,s in enumerate(cut):
            #    s.base.exportStep('cutplane_{}.stp'.format(i))
            # print('planes',full)
            # print('planes',cut)
        else:
            full = []
            cut = base

        if others:
            newf, cut = SplitSolid(cut, others, cell, tolerance=Options.splitTolerance)
            # print('others',newf)
            # print('others',cut)
        else:
            newf = []
        full.extend(newf)

    elif mode == "otherFirst":
        planes = []
        others = []
        for s in surfaces:
            # s.buildShape( boundBox)
            if s.type == "plane":
                planes.append(s)
            else:
                others.append(s)

        if others:
            full, cut = SplitSolid(base, others, cell, tolerance=Options.splitTolerance)
            # print('others',full)
            # print('others',cut)
        else:
            full = []
            cut = base

        if planes:
            newf, cut = SplitSolid(cut, planes, cell, tolerance=Options.splitTolerance)
            # print('planes',newf)
            # print('planes',cut)

        else:
            newf = []

        full.extend(newf)

    elif mode == "oneByOne":
        planes = []
        others = []
        for s in surfaces:
            if s.type == "plane":
                planes.append(s)
            else:
                others.append(s)

        if planes:
            full, cut = SplitSolid(base, planes, cell, tolerance=Options.splitTolerance)
        else:
            full = []
            cut = base

        # cut[0].base.exportStep('cutPlane.stp')
        for surf in others:
            newf, cut = SplitSolid(cut, (surf,), cell, tolerance=Options.splitTolerance)
            full.extend(newf)

    elif mode == "otherOneByOne":
        planes = []
        others = []
        for s in surfaces:
            if s.type == "plane":
                planes.append(s)
            else:
                others.append(s)

        cut = base
        full = []
        for surf in others:
            newf, cut = SplitSolid(cut, (surf,), cell, tolerance=Options.splitTolerance)
            full.extend(newf)

        for surf in planes:
            newf, cut = SplitSolid(cut, (surf,), cell, tolerance=Options.splitTolerance)
            full.extend(newf)

    return full, cut


def FuseSolid(parts):
    if (len(parts)) <= 1:
        if parts:
            solid = parts[0]
        else:
            return None
    else:
        try:
            fused = parts[0].fuse(parts[1:])
        except:
            fused = None

        if fused is not None:
            try:
                refinedfused = fused.removeSplitter()
            except:
                refinedfused = fused

            if refinedfused.isValid():
                solid = refinedfused
            else:
                if fused.isValid():
                    solid = fused
                else:
                    solid = Part.makeCompound(parts)
        else:
            solid = Part.makeCompound(parts)

    if solid.Volume < 0:
        solid.reverse()
    return solid


def noOR(Seq):
    if len(Seq.elements) == 1:
        # Seq.operator = 'AND'
        return Seq
    newOR = BoolSequence(operator="OR")
    neg = []
    for e in Seq.elements:
        AND = BoolSequence(operator="AND")
        AND.append(*neg, e)
        newOR.append(AND)
        neg.append(-e)
    return newOR



#################################
# Module to write Serpent input #
#################################
import logging
from datetime import datetime
from pathlib import Path
from importlib.metadata import version

import FreeCAD

from ..utils.basic_functions_part1 import is_opposite, points_to_coeffs
from ..utils.functions import SurfacesDict
from .functions import change_surf_sign, serpent_surface, write_serpent_cell_def

logger = logging.getLogger("general_logger")


class SerpentInput:
    def __init__(
        self,
        Meta,
        Surfaces,
        options,
        tolerances,
        numeric_format,
        title,
        volSDEF,
        volCARD,
        UCARD,
        dummyMat,
        step_filename,
    ):
        self.options = options
        self.tolerances = tolerances
        self.numeric_format = numeric_format
        self.Title = title
        self.VolSDEF = volSDEF
        self.VolCARD = volCARD
        self.U0CARD = UCARD
        self.dummyMat = dummyMat
        self.Cells = Meta
        self.Options = {
            "Volume": self.VolCARD,
            "Particle": ("n", "p"),
            "Universe": self.U0CARD,
        }
        self.part = "p"

        self.step_filename = step_filename
        if isinstance(self.step_filename, (tuple, list)):
            self.step_filename = "; ".join(self.step_filename)

        self.get_surface_table()
        self.simplify_planes(Surfaces)

        self.Surfaces = self.sorted_surfaces(Surfaces)
        self.Materials = set()

        return

    # def setSRC(self,data):

    #     if data[0] is not None:
    #       sdef     = f'src point {self.part}\n'
    #       self.src_sphere = (sdef)
    #     else:
    #       self.src_sphere = None
    #     xmin,xmax,ymin,ymax,zmin,zmax = data[1]

    #     sdef = 'SDEF PAR={} X=D1 Y=D2 Z=D3 \n'.format(self.part)
    #     SI1  = 'SI1 {:13.7e} {:13.7e} \n'.format(xmin*0.1,xmax*0.1)
    #     SI2  = 'SI2 {:13.7e} {:13.7e} \n'.format(ymin*0.1,ymax*0.1)
    #     SI3  = 'SI3 {:13.7e} {:13.7e} \n'.format(zmin*0.1,zmax*0.1)
    #     SP1  = 'SP1 0  1 \n'
    #     SP2  = 'SP2 0  1 \n'
    #     SP3  = 'SP3 0  1 \n'
    #     self.SDEF_box    = (sdef,SI1,SI2,SI3,SP1,SP2,SP3)

    def write_input(self, filename):
        logger.info(f"write Serpent file {filename}")
        Path(filename).parent.mkdir(parents=True, exist_ok=True)
        with open(file=filename, mode="w", encoding="utf-8") as self.inpfile:
            self.write_header()
            cellblockHeader = """\
% --- CELL DEFINITIONS 
"""
            self.inpfile.write(cellblockHeader)
            self.write_cell_block()
            self.inpfile.write(" \n")

            surfaceHeader = """\
% --- SURFACE DEFINITIONS 
"""
            self.inpfile.write(surfaceHeader)
            self.write_surface_block()
            self.inpfile.write(" \n")

            self.write_source_block()

        return

    def write_header(self):

        freeCAD_Version = "{V[0]:}.{V[1]:}.{V[2]:}".format(V=FreeCAD.Version())

        Header = f"""{self.Title}
%   ______ _______  _____      _     _ __   _ _______ ______  
%  |  ____ |______ |     | ___ |     | | \\  | |______ |     \\ 
%  |_____| |______ |_____|     |_____| |  \\_| |______ |_____/
% Version : {version('geouned')}
% FreeCAD Version : {freeCAD_Version} 
"""

        Information = f"""%
% *************************************************************
% Original Step file : {self.step_filename}
%
% Creation Date : {datetime.now()}
% Solid Cells   : {self.__solidCells__}
% Total Cells   : {self.__cells__}
% Surfaces      : {len(self.Surfaces)}
% Materials     : {len(self.__materials__)}
%
% **************************************************************
"""
        self.inpfile.write(Header)
        self.inpfile.write(Information)
        return

    def write_surface_block(self):

        for surf in self.Surfaces:
            self.write_surfaces(surf)

    def write_cell_block(self):

        for i, cell in enumerate(self.Cells):
            self.write_cells(cell)
        return

    # to_mod
    def write_cells(self, cell):
        index = cell.label

        # If index is None, the object does not contain cell definition
        # but a comment to insert between cells
        if cell.__id__ is None:
            comment = self.comment_line(cell.Comments)
            self.inpfile.write(comment)
            return

        if self.Options["Universe"] is not None:
            if cell.Material == 0:
                cellHeader = (
                    # {"void":<5d} has been removed from the end of the line below
                    # see issue https://github.com/GEOUNED-org/GEOUNED/issues/151 for details
                    f'cell {index:<5d} {self.Options["Universe"]} '
                )
            else:
                self.Materials.add(cell.Material)
                cellHeader = f'cell {index:<5d} {self.Options["Universe"]} {cell.Material:<5d} '

            serpent_cell = (
                f"{cellHeader}{self.cell_format(cell.Definition, offset=len(cellHeader))}"
                f"{self.comment_format(cell.Comments, cell.MatInfo)}"
            )
            self.inpfile.write(serpent_cell)
        else:
            if cell.Material == 0 and cell.MatInfo != "Graveyard":
                cellHeader = f"cell {index:<5d} 0 void "
            elif cell.Material == 0 and cell.MatInfo == "Graveyard":
                cellHeader = f"cell {index:<5d} 0 outside "
            else:
                self.Materials.add(cell.Material)
                cellHeader = f"cell {index:<5d}  0  {cell.Material:<5d} "

        serpent_cell = (
            f"{cellHeader}{self.cell_format(cell.Definition, offset=len(cellHeader))}"
            f"{self.comment_format(cell.Comments, cell.MatInfo)}"
        )
        self.inpfile.write(serpent_cell)

        return

    def write_surfaces(self, surface):
        """Write the surfaces in Serpent format"""

        Serpent_def = serpent_surface(
            surface.Index,
            surface.Type,
            surface.Surf,
            self.options,
            self.tolerances,
            self.numeric_format,
        )
        if Serpent_def:
            Serpent_def += "\n"
            self.inpfile.write(Serpent_def)
        else:
            logger.info(f"Surface {surface.Type} cannot be written in Serpent input")
        return

    # No void all option in Serpent. For now remove addition of source.

    def write_source_block(self):

        #       if self.SDEF_sphere is None:  return
        MODE = f"\nset nps 1e6\nset bc 1"
        if self.dummyMat:
            mat = list(self.Materials)
            mat.sort()
            MATCARD = ""
            for m in mat:
                MATCARD += f"mat {m:<6d} {self.cell.Density:11.4e} \n1001 1 \n"
            Block = MATCARD + "% \n" + MODE
        else:
            Block = MODE

        # Not included for Serpent
        #    if self.VolSDEF:
        #       Block += 'PRDMP 2J -1\n'
        #       for line in self.SDEF_box :
        #          Block += 'C ' + line
        #       for line in self.SDEF_sphere:
        #          Block += line

        #       celList,volList = self.get_solid_cell_volume()

        #       F4Tally = CardLine('F4:{} '.format(self.part))
        #       F4Tally.extend(celList)
        #       SD4     = CardLine('SD4  ',fmt='13.7e')
        #       SD4.extend(volList)

        #       Block += F4Tally.str+'\n'
        #       if not self.VolCARD :
        #          Block += SD4.str
        #       else :
        #          Block += 'C Cell volume normalization is set in cell cards VOL\n'
        #    else :
        #       for line in self.SDEF_sphere :
        #          Block += 'C ' + line
        #       for line in self.SDEF_box :
        #          Block +=  line

        self.inpfile.write(Block)

    def cell_format(self, Definition, offset=11):
        return write_serpent_cell_def(Definition, tabspace=11, offset=offset)

    # Function not relevant for Serpent : No importance setting, universes assigned elsewhere.
    # Volumes only defined on tally cards.
    # def option_format(self,cell):

    #    option = ''
    #    if self.Options['Volume']:
    #       if not cell.Void :
    #         option ='{:11s}Vol={:e}\n'.format('',cell.Volume*1e-3)
    #       else:
    #         option = '{:11s}Vol=1.0\n'.format('')

    #    option +=  '{:11s}'.format('')
    #    for p in self.Options['Particle']:
    #      if cell.MatInfo == 'Graveyard' :
    #        option +=  'imp:{}=0     '.format(p)
    #      else:
    #        option +=  'imp:{}=1.0   '.format(p)

    #    if self.Options['Universe'] is not None:
    #         option += 'U={}  '.format(self.Options['Universe'])
    #    option += '\n'

    #    return option

    def comment_format(self, cComment, mComment=None):
        comment = ""
        if mComment:
            mComment = mComment.split("\n")
            for c in mComment:
                if c:
                    comment += f"{'':11s}%{c}\n"

        if cComment.strip() != "":
            cComment = cComment.strip().split("\n")
            for c in cComment:
                if c:
                    comment += f"{'':11s}%{c}\n"
        return comment

    def comment_line(self, lineComment):
        lineComment = lineComment.strip().split("\n")
        comment = ""
        if lineComment:
            comment = "% \n"
            for c in lineComment:
                if c:
                    comment += f"% {c}\n"
            comment += "% \n"
        return comment

    def get_surface_table(self):
        self.surfaceTable = {}
        self.__solidCells__ = 0
        self.__cells__ = 0
        self.__materials__ = set()

        for i, CellObj in enumerate(self.Cells):
            if CellObj.__id__ is None:
                continue
            self.__cells__ += 1
            if CellObj.Material != 0:
                self.__materials__.add(CellObj.Material)

            surf = CellObj.Definition.get_surfaces_numbers()
            if not CellObj.Void:
                self.__solidCells__ += 1
            for index in surf:
                if index in self.surfaceTable.keys():
                    self.surfaceTable[index].add(i)
                else:
                    self.surfaceTable[index] = {i}
        return

    def simplify_planes(self, Surfaces):

        for p in Surfaces["PX"]:
            if p.Surf.Axis[0] < 0:
                p.Surf.Axis = FreeCAD.Vector(1, 0, 0)
                self.change_surf_sign(p)

        for p in Surfaces["PY"]:
            if p.Surf.Axis[1] < 0:
                p.Surf.Axis = FreeCAD.Vector(0, 1, 0)
                self.change_surf_sign(p)

        for p in Surfaces["PZ"]:
            if p.Surf.Axis[2] < 0:
                p.Surf.Axis = FreeCAD.Vector(0, 0, 1)
                self.change_surf_sign(p)

        if self.options.prnt3PPlane:
            for p in Surfaces["P"]:
                if p.Surf.pointDef:
                    axis, d = points_to_coeffs(p.Surf.Points)
                    if is_opposite(axis, p.Surf.Axis):
                        self.change_surf_sign(p)

        return

    def sorted_surfaces(self, Surfaces):
        temp = SurfacesDict(Surfaces)
        surfList = []
        for ind in range(Surfaces.IndexOffset, Surfaces.surfaceNumber + Surfaces.IndexOffset):
            s = temp.get_surface(ind + 1)
            if s is not None:
                surfList.append(s)
                temp.del_surface(ind + 1)
        return surfList

    def change_surf_sign(self, p):

        if p.Index not in self.surfaceTable.keys():
            logger.info(f"{p.Type} Surface {p.Index} not used in cell definition) {p.Surf.Axis} {p.Surf.Position}")
            return

        for ic in self.surfaceTable[p.Index]:
            surf = self.Cells[ic].Definition.get_surfaces_numbers()
            for s in surf:
                if s == p.Index:
                    change_surf_sign(s, self.Cells[ic].Definition)

    def get_solid_cell_volume(self):

        solidList = []
        volumeList = []
        for m in self.Cells:
            if m.CellType == "solid" and m.__id__ is not None:
                solidList.append(m.label)
                volumeList.append(m.Volume * 1e-3)

        return solidList, volumeList



#
# Set of useful functions used in different parts of the code
#
import math

import FreeCAD


def is_same_value(v1, v2, tolerance=1e-6):
    return abs(v1 - v2) < tolerance


def is_opposite(vector_1, vector_2, tolerance=1e-6):
    return abs(vector_1.getAngle(-vector_2)) < tolerance


def is_parallel(vector_1, vector_2, tolerance=1e-6):
    angle = abs(vector_1.getAngle(vector_2))
    return angle < tolerance or is_same_value(angle, math.pi, tolerance)


def is_in_line(point, dir, pnt_line, tolerance=1e-6):
    r12 = point - pnt_line
    return is_parallel(dir, r12) or (r12.Length < tolerance)


# TODO check this function is used in the code
def is_in_points(point, points, tolerance=1e-5):
    if len(points) > 0:
        for p in points:
            if point.isEqual(p, tolerance):
                return True
    return False


# TODO check this function is used in the code
def is_in_edge(edge1, edge2, tolerance=1e-8):
    ver1 = edge1.Vertexes
    ver2 = edge2.Vertexes
    con1 = ver1[0].Point.isEqual(ver2[0].Point, tolerance) or ver1[0].Point.isEqual(ver2[1].Point, tolerance)
    con2 = ver1[1].Point.isEqual(ver2[0].Point, tolerance) or ver1[1].Point.isEqual(ver2[1].Point, tolerance)
    return con1 and con2


def is_in_plane(point, plane, d_tolerance=1e-7):
    return abs(point.distanceToPlane(plane.Surf.Position, plane.Surf.Axis)) < d_tolerance


def is_in_tolerance(val, tol, fuzzy_low, fuzzy_high):
    if abs(val) < fuzzy_low:
        return True, False  # 1) isintolerance 2) fuzzy
    elif abs(val) < tol:
        return True, True
    elif abs(val) > fuzzy_high:
        return False, False
    else:
        return False, True


def sign_plane(point, plane):
    value = plane.Surf.Axis.dot(point - plane.Surf.Position)
    if value >= 0.0:
        sign = 1
    else:
        sign = -1
    return sign


def points_to_coeffs(points):
    p1, p2, p3 = points[0:3]
    scf = (p1.x, p1.y, p1.z, p2.x, p2.y, p2.z, p3.x, p3.y, p3.z)

    # mcnp implementation to convert 3 point plane to
    # plane parameters

    tpp = [0] * 4
    for i in range(1, 4):
        j = i % 3 + 1
        k = 6 - i - j
        k -= 1
        j -= 1
        tpp[i - 1] = (
            scf[j] * (scf[k + 3] - scf[k + 6]) + scf[j + 3] * (scf[k + 6] - scf[k]) + scf[j + 6] * (scf[k] - scf[k + 3])
        )
        tpp[3] += scf[i - 1] * (scf[j + 3] * scf[k + 6] - scf[j + 6] * scf[k + 3])

    xm = 0
    coeff = [0] * 4
    for i in range(1, 5):
        if xm == 0 and tpp[4 - i] != 0:
            xm = 1 / tpp[4 - i]
        coeff[4 - i] = tpp[4 - i] * xm

    axis = FreeCAD.Vector(coeff[0:3])
    distance = coeff[3] / axis.Length
    axis.normalize()

    return axis, distance


class Plane3PtsParams:
    def __init__(self, params, real=True):
        self.Position = params[0]
        self.Axis = params[1]
        self.dimL1 = params[2]
        self.dimL2 = params[3]
        self.Points = params[4]
        self.real = real
        self.pointDef = True

    def __str__(self):
        #      outstr = '''Plane :
        #    Point 1  : {P1[0]}  {P1[1]}  {P1[2]}
        #    Point 2  : {P2[0]}  {P2[1]}  {P2[2]}
        #    Point 3  : {P3[0]}  {P3[1]}  {P3[2]} '''.format(P1=self.Points[0], P2=self.Points[1], P3=self.Points[2])
        pos = self.Axis.dot(self.Position)
        outstr = f"""Plane :
    Axis     : {self.Axis.x}  {self.Axis.y}  {self.Axis.z} 
    Position : {pos}  """
        return outstr


class PlaneParams:
    def __init__(self, params, real=True):
        self.Position = params[0]
        self.Axis = params[1]
        self.dimL1 = params[2]
        self.dimL2 = params[3]
        self.real = real
        self.pointDef = False

    def __str__(self):
        pos = self.Axis.dot(self.Position)
        outstr = f"""Plane :
    Axis     : {self.Axis.x}  {self.Axis.y}  {self.Axis.z} 
    Position : {pos}  """
        return outstr


class CylinderParams:
    def __init__(self, params, real=True):
        self.Center = params[0]
        self.Axis = params[1]
        self.Radius = params[2]
        self.dimL = params[3]
        self.real = real

    def __str__(self):
        outstr = f"""Cylinder :
    Axis     : {self.Axis.x}  {self.Axis.y}  {self.Axis.z} 
    Center   : {self.Center.x}  {self.Center.y}  {self.Center.z}
    Radius   : {self.Radius}  """
        return outstr


class ConeParams:
    def __init__(self, params, real=True):
        self.Apex = params[0]
        self.Axis = params[1]
        self.SemiAngle = params[2]
        self.dimL = params[3]
        self.dimR = params[4]
        self.real = real

    def __str__(self):
        outstr = f"""Cone :
    Axis     : {self.Axis.x}  {self.Axis.y}  {self.Axis.z} 
    Center   : {self.Apex.x}  {self.Apex.y}  {self.Apex.z}
    SemiAngle: {self.SemiAngle}  """
        return outstr


class SphereParams:
    def __init__(self, params):
        self.Center = params[0]
        self.Radius = params[1]

    def __str__(self):
        outstr = f"""Sphere :
    Center   : {self.Center.x}  {self.Center.y}  {self.Center.z}
    Radius   : {self.Radius}  """
        return outstr


class TorusParams:
    def __init__(self, params):
        self.Center = params[0]
        self.Axis = params[1]
        self.MajorRadius = params[2]
        self.MinorRadius = params[3]

    def __str__(self):
        outstr = f"""Torus :
    Axis     : {self.Axis.x}  {self.Axis.y}  {self.Axis.z} 
    Center   : {self.Center.x}  {self.Center.y}  {self.Center.z}
    MajorRadius: {self.MajorRadius}
    MinorRadius: {self.MinorRadius} """
        return outstr



from pathlib import Path
import pytest
import geouned


@pytest.mark.parametrize("csg_format", ["mcnp", "openmc_xml"])
def test_cylbox_convertion(csg_format):

    if csg_format == "openmc_xml":
        suffix = ".xml"
    elif csg_format == "mcnp":
        suffix = ".mcnp"

    geo = geouned.CsgToCad()

    geo.export_cad(
        # csg file was made from testing/inputSTEP/cylBox.stp
        input_filename=f"tests/csg_files/cylinder_box{suffix}",
        csg_format=csg_format,
        bounding_box=[-1000.0, -500.0, -1000.0, 0, 0, 0.0],
        # TODO add tests for these args that counts volumes in cad file
        # cell_range_type='exclude',
        # cell_range=(2,3,4),
        output_filename=f"tests_outputs/csgtocad/{csg_format}",
    )

    assert Path(f"tests_outputs/csgtocad/{csg_format}.step").exists()
    assert Path(f"tests_outputs/csgtocad/{csg_format}.FCStd").exists()



import json
import logging
import typing
from datetime import datetime
from pathlib import Path
from typing import get_type_hints
from importlib.metadata import version

import FreeCAD
import Part
from tqdm import tqdm

from .code_version import *
from .conversion import cell_definition as Conv
from .cuboid.translate import translate
from .decompose import decom_one as Decom
from .loadfile import load_step as Load
from .utils import functions as UF
from .utils.boolean_solids import build_c_table_from_solids
from .utils.data_classes import NumericFormat, Options, Settings, Tolerances
from .void import void as void
from .write.functions import write_mcnp_cell_def
from .write.write_files import write_geometry

logger = logging.getLogger("general_logger")
logger.info(f"GEOUNED version {version('geouned')}")
logger.info(f"FreeCAD version {'.'.join(FreeCAD.Version()[:3])}")


class CadToCsg:
    """Base class for the conversion of CAD to CSG models

    Args:
        options (geouned.Options, optional): An instance of a geouned.Options
            class with the attributes set for the desired conversion. Defaults
            to a geouned.Options with default attributes values.
        tolerances (geouned.Tolerances, optional): An instance of a
            geouned.Tolerances class with the attributes set for the desired
            conversion. Defaults to a geouned.Tolerances with default
            attributes values.
        numeric_format (geouned.NumericFormat, optional): An instance of a
            geouned.NumericFormat class with the attributes set for the desired
            conversion. Defaults to a geouned.NumericFormat with default
            attributes values.
        settings (geouned.Settings, optional): An instance of a
            geouned.Settings class with the attributes set for the desired
            conversion. Defaults to a geouned.Settings with default
            attributes values.
    """

    def __init__(
        self,
        options: Options = Options(),
        tolerances: Tolerances = Tolerances(),
        numeric_format: NumericFormat = NumericFormat(),
        settings: Settings = Settings(),
    ):

        self.options = options
        self.tolerances = tolerances
        self.numeric_format = numeric_format
        self.settings = settings

        # define later when running the code
        self.geometry_bounding_box = None
        self.meta_list = []
        self.filename = None
        self.skip_solids = []

    @property
    def options(self):
        return self._options

    @options.setter
    def options(self, value: Options):
        if not isinstance(value, Options):
            raise TypeError(f"geouned.CadToCsg.options should be an instance of geouned.Options, not a {type(value)}")
        self._options = value

    @property
    def tolerances(self):
        return self._tolerances

    @tolerances.setter
    def tolerances(self, value: tolerances):
        if not isinstance(value, Tolerances):
            raise TypeError(f"geouned.CadToCsg.tolerances should be an instance of geouned.Tolerances, not a {type(value)}")
        self._tolerances = value

    @property
    def numeric_format(self):
        return self._numeric_format

    @numeric_format.setter
    def numeric_format(self, value: numeric_format):
        if not isinstance(value, NumericFormat):
            raise TypeError(
                f"geouned.CadToCsg.numeric_format should be an instance of geouned.NumericFormat, not a {type(value)}"
            )
        self._numeric_format = value

    @property
    def settings(self):
        return self._settings

    @settings.setter
    def settings(self, value: settings):
        if not isinstance(value, Settings):
            raise TypeError(f"geouned.CadToCsg.settings should be an instance of geouned.Settings, not a {type(value)}")
        self._settings = value

    def export_csg(
        self,
        title: str = "Converted with GEOUNED",
        geometryName: str = "csg",
        outFormat: typing.Tuple[str] = (
            "openmc_xml",
            "openmc_py",
            "serpent",
            "phits",
            "mcnp",
        ),
        volSDEF: bool = False,
        volCARD: bool = True,
        UCARD: typing.Union[int, None] = None,
        dummyMat: bool = False,
        cellCommentFile: bool = False,
        cellSummaryFile: bool = True,
    ):
        """Writes out a CSG file in the requested Monte Carlo code format.

        Args:
            title (str, optional): Title of the model written at the top of the
                output file. Defaults to "Geouned conversion".
            geometryName (str, optional): the file stem of the output file(s).
                Defaults to "converted_with_geouned".
            outFormat (typing.Tuple[str], optional): Format for the output
                geometry. Available format are: "mcnp", "openmc_xml",
                "openmc_py", "phits" and "serpent". Several output format can
                be written in the same method call. Defaults to output all codes.
            volSDEF (bool, optional):  Write SDEF definition and tally of solid
                cell for stochastic volume checking. Defaults to False.
            volCARD (bool, optional): Write the CAD calculated volume in the
                cell definition using the VOL card. Defaults to True.
            UCARD (int, optional): Write universe card in the cell definition
                with the specified universe number (if value = 0 Universe card
                is not written). Defaults to None.
            dummyMat (bool, optional): Write dummy material definition card in
               the MCNP output file for all material labels present in the
               model. Dummy material definition is "MX 1001 1". Defaults to False.
            cellCommentFile (bool, optional): Write an additional file with
               comment associated to each CAD cell in the MCNP output file.
               Defaults to False.
            cellSummaryFile (bool, optional): Write an additional file with
               information on the CAD cell translated. Defaults to True.
        """

        if not isinstance(UCARD, int) and not isinstance(UCARD, type(None)):
            raise TypeError(f"UCARD should be of type int or None not {type(UCARD)}")
        if isinstance(UCARD, int):
            if UCARD < 0:
                raise ValueError("UCARD should be a 0 or a positive integer ")

        for arg, arg_str in (
            (volSDEF, "volSDEF"),
            (volCARD, "volCARD"),
            (dummyMat, "dummyMat"),
            (cellCommentFile, "cellCommentFile"),
            (cellSummaryFile, "cellSummaryFile"),
        ):
            if not isinstance(arg, bool):
                raise TypeError(f"{arg} should be of type bool not {type(arg_str)}")

        for arg, arg_str in ((title, "title"), (geometryName, "geometryName")):
            if not isinstance(arg, str):
                raise TypeError(f"{arg} should be of type str not {type(arg_str)}")

        # if the geometry_bounding_box has not previuosly been calculated, then make a default one
        if self.geometry_bounding_box is None:
            self._get_geometry_bounding_box()

        write_geometry(
            UniverseBox=self.geometry_bounding_box,
            MetaList=self.meta_list,
            Surfaces=self.Surfaces,
            settings=self.settings,
            options=self.options,
            tolerances=self.tolerances,
            numeric_format=self.numeric_format,
            geometryName=geometryName,
            outFormat=outFormat,
            cellCommentFile=cellCommentFile,
            cellSummaryFile=cellSummaryFile,
            title=title,
            volSDEF=volSDEF,
            volCARD=volCARD,
            UCARD=UCARD,
            dummyMat=dummyMat,
            step_filename=self.filename,
        )

        logger.info("End of Monte Carlo code translation phase")

    @classmethod
    def from_json(cls, filename: str):
        """Creates a CadToCsg instance, runs CadToCsg.load_Step_file(), runs
        CadToCsg.start() and returns the instance. Populating the arguments for
        the methods that are run by looking for keys with the same name as the
        method in the JSON file. For example CadToCsg.start() accepts arguments
        for Options, Tolerance, Settings and NumericFormat and can be populated
        from matching key names in the JSON. If export_to_csg key is present
        then this method also runs CadToCsg.export_to_csg() on the instance.

        Args:
            filename str: The filename of the config file.

        Raises:
            FileNotFoundError: If the config file is not found
            ValueError: If the config JSON file is found to contain an invalid key

        Returns:
            geouned.CadToCsg: returns a geouned CadToCsg class instance.
        """

        if not Path(filename).exists():
            raise FileNotFoundError(f"config file {filename} not found")

        with open(filename) as f:
            config = json.load(f)

        cad_to_csg = cls()

        for key in config.keys():

            if key in ["load_step_file", "export_csg"]:
                pass  # these two keys are used before or after this loop

            elif key == "Tolerances":
                cad_to_csg.tolerances = Tolerances(**config["Tolerances"])

            elif key == "Options":
                cad_to_csg.options = Options(**config["Options"])

            elif key == "NumericFormat":
                cad_to_csg.numeric_format = NumericFormat(**config["NumericFormat"])

            elif key == "Settings":
                cad_to_csg.settings = Settings(**config["Settings"])

            else:
                raise ValueError(
                    f"Invalid key '{key}' found in config file {filename}. Acceptable key names are 'load_step_file', 'export_csg', 'Settings', 'Parameters', 'Tolerances' and 'NumericFormat'"
                )

        cad_to_csg.load_step_file(**config["load_step_file"])
        cad_to_csg.start()
        if "export_csg" in config.keys():
            cad_to_csg.export_csg(**config["export_csg"])
        else:
            cad_to_csg.export_csg()
        return cad_to_csg

    def load_step_file(
        self,
        filename: typing.Union[str, typing.Sequence[str]],
        skip_solids: typing.Sequence[int] = [],
    ):
        """
        Load STEP file(s) and extract solid volumes and enclosure volumes.

        Args:
            filename (str): The path to the STEP file or a list of paths to multiple STEP files.
            skip_solids (Sequence[int], optional): A sequence (list or tuple) of indexes of solids to not load for conversion.

        Returns:
            tuple: A tuple containing the solid volumes list and enclosure volumes list extracted from the STEP files.
        """
        logger.info("Start of step file loading phase")

        if not isinstance(skip_solids, (list, tuple)):
            raise TypeError(f"skip_solids should be a list, tuple of ints, not a {type(skip_solids)}")
        for entry in skip_solids:
            if not isinstance(entry, int):
                raise TypeError(f"skip_solids should contain only ints, not a {type(entry)}")

        if not isinstance(filename, (str, list, tuple)):
            raise TypeError(f"filename should be a str or a sequence of str, not a {type(filename)}")
        if isinstance(filename, (list, tuple)):
            for entry in filename:
                if not isinstance(entry, str):
                    raise TypeError(f"filename should contain only str, not a {type(entry)}")

        self.filename = filename
        self.skip_solids = skip_solids

        if isinstance(filename, (list, tuple)):
            step_files = filename
        else:
            step_files = [filename]

        for step_file in step_files:
            if not Path(step_file).is_file():
                raise FileNotFoundError(f"Step file {step_file} not found.")

        MetaChunk = []
        EnclosureChunk = []
        for step_file in tqdm(step_files, desc="Loading CAD files"):
            logger.info(f"read step file : {step_file}")
            Meta, Enclosure = Load.load_cad(step_file, self.settings, self.options)
            MetaChunk.append(Meta)
            EnclosureChunk.append(Enclosure)
        self.meta_list = join_meta_lists(MetaChunk)
        self.enclosure_list = join_meta_lists(EnclosureChunk)

        # deleting the solid index in reverse order so the indexes don't change for subsequent deletions
        for solid_index in sorted(skip_solids, reverse=True):
            logger.info(f"Removing solid index: {solid_index} from list of {len(self.meta_list)} solids")
            del self.meta_list[solid_index]

        logger.info("End of step file loading phase")

        return self.meta_list, self.enclosure_list

    def _export_solids(self, filename: str):
        """Export all the solid volumes from the loaded geometry to a STEP file.

        Args:
            filename (str): filepath of the output STEP file.
        """
        # export in STEP format solids read from input file
        if self.meta_list == []:
            raise ValueError(
                "No solids in CadToCsg.meta_list to export. Try loading the STEP file first with CadToCsg.load_step_file"
            )
        solids = []
        for m in self.meta_list:
            if m.IsEnclosure:
                continue
            solids.extend(m.Solids)
        Part.makeCompound(solids).exportStep(filename)

    def _get_geometry_bounding_box(self, padding: float = 10.0):
        """
        Get the bounding box of the geometry.

        Args:
            padding (float): The padding value to add to the bounding box dimensions.

        Returns:
            FreeCAD.BoundBox: The universe bounding box.
        """
        # set up Universe
        meta_list = self.meta_list

        Box = meta_list[0].optimalBoundingBox()
        xmin = Box.XMin
        xmax = Box.XMax
        ymin = Box.YMin
        ymax = Box.YMax
        zmin = Box.ZMin
        zmax = Box.ZMax
        for m in meta_list[1:]:
            # MIO. This was removed since in HELIAS the enclosure cell is the biggest one
            # if m.IsEnclosure: continue
            optBox = m.optimalBoundingBox()
            xmin = min(optBox.XMin, xmin)
            xmax = max(optBox.XMax, xmax)
            ymin = min(optBox.YMin, ymin)
            ymax = max(optBox.YMax, ymax)
            zmin = min(optBox.ZMin, zmin)
            zmax = max(optBox.ZMax, zmax)

        self.geometry_bounding_box = FreeCAD.BoundBox(
            FreeCAD.Vector(xmin - padding, ymin - padding, zmin - padding),
            FreeCAD.Vector(xmax + padding, ymax + padding, zmax + padding),
        )
        return self.geometry_bounding_box

    def start(self):

        startTime = datetime.now()

        if self.settings.exportSolids:
            self._export_solids(filename=self.settings.exportSolids)

        logger.info("End of loading phase")
        tempstr1 = str(datetime.now() - startTime)
        logger.info(tempstr1)
        tempTime = datetime.now()

        # sets self.geometry_bounding_box with default padding
        self._get_geometry_bounding_box()

        self.Surfaces = UF.SurfacesDict(offset=self.settings.startSurf - 1)

        warnSolids = []
        warnEnclosures = []
        coneInfo = dict()
        tempTime0 = datetime.now()
        if not self.options.Facets:

            # decompose all solids in elementary solids (convex ones)
            warningSolidList = self._decompose_solids(meta=True)

            # decompose Enclosure solids
            if self.settings.voidGen and self.enclosure_list:
                warningEnclosureList = self._decompose_solids(meta=False)

            logger.info("End of decomposition phase")

            # start Building CGS cells phase

            for j, m in enumerate(tqdm(self.meta_list, desc="Translating solid cells")):
                if m.IsEnclosure:
                    continue
                logger.info(f"Building cell: {j+1}")
                cones = Conv.cellDef(
                    m,
                    self.Surfaces,
                    self.geometry_bounding_box,
                    self.options,
                    self.tolerances,
                    self.numeric_format,
                )
                if cones:
                    coneInfo[m.__id__] = cones
                if j in warningSolidList:
                    warnSolids.append(m)
                if not m.Solids:
                    logger.info(f"none {j}, {m.__id__}")
                    logger.info(m.Definition)

            if self.options.forceNoOverlap:
                Conv.no_overlapping_cell(self.meta_list, self.Surfaces, self.options)

        else:
            translate(
                self.meta_list,
                self.Surfaces,
                self.geometry_bounding_box,
                self.settings,
                self.options,
                self.tolerances,
            )
            # decompose Enclosure solids
            if self.settings.voidGen and self.enclosure_list:
                warningEnclosureList = self._decompose_solids(meta=False)

        tempstr2 = str(datetime.now() - tempTime)
        logger.info(tempstr2)

        #  building enclosure solids

        if self.settings.voidGen and self.enclosure_list:
            for j, m in enumerate(self.enclosure_list):
                logger.info(f"Building Enclosure Cell: {j + 1}")
                cones = Conv.cellDef(
                    m,
                    self.Surfaces,
                    self.geometry_bounding_box,
                    self.options,
                    self.tolerances,
                    self.numeric_format,
                )
                if cones:
                    coneInfo[m.__id__] = cones
                if j in warningEnclosureList:
                    warnEnclosures.append(m)

        tempTime1 = datetime.now()

        # void generation phase
        meta_void = []
        if self.settings.voidGen:
            logger.info("Build Void")
            logger.info(self.settings.voidExclude)
            if not self.settings.voidExclude:
                meta_reduced = self.meta_list
            else:
                meta_reduced = exclude_cells(self.meta_list, self.settings.voidExclude)

            if self.meta_list:
                init = self.meta_list[-1].__id__ - len(self.enclosure_list)
            else:
                init = 0
            meta_void = void.void_generation(
                meta_reduced,
                self.enclosure_list,
                self.Surfaces,
                self.geometry_bounding_box,
                self.settings,
                init,
                self.options,
                self.tolerances,
                self.numeric_format,
            )

        # if self.settings.simplify == 'full' and not self.options.forceNoOverlap:
        if self.settings.simplify == "full":
            Surfs = {}
            for lst in self.Surfaces.values():
                for s in lst:
                    Surfs[s.Index] = s

            for c in tqdm(self.meta_list, desc="Simplifying"):
                if c.Definition.level == 0 or c.IsEnclosure:
                    continue
                logger.info(f"simplify cell {c.__id__}")
                Box = UF.get_box(c, self.options.enlargeBox)
                CT = build_c_table_from_solids(Box, (c.Surfaces, Surfs), "full", options=self.options)
                c.Definition.simplify(CT)
                c.Definition.clean()
                if type(c.Definition.elements) is bool:
                    logger.info(f"unexpected constant cell {c.__id__} :{c.Definition.elements}")

        tempTime2 = datetime.now()
        logger.info(f"build Time: {tempTime2} - {tempTime1}")

        logger.info(datetime.now() - startTime)

        cellOffSet = self.settings.startCell - 1
        if self.enclosure_list and self.settings.sort_enclosure:
            # sort group solid cell / void cell sequence in each for each enclosure
            # if a solid belong to several enclosure, its definition will be written
            # for the highest enclosure level or if same enclosure level in the first
            # enclosure found
            self.meta_list = sort_enclosure(self.meta_list, meta_void, cellOffSet)
        else:
            # remove Null Cell and apply cell numbering offset
            deleted = []
            idLabel = {0: 0}
            icount = cellOffSet
            for i, m in enumerate(self.meta_list):
                if m.NullCell or m.IsEnclosure:
                    deleted.append(i)
                    continue

                icount += 1
                m.label = icount
                idLabel[m.__id__] = m.label

            for i in reversed(deleted):
                del self.meta_list[i]

            lineComment = """\
##########################################################
             VOID CELLS
##########################################################"""
            mc = UF.GeounedSolid(None)
            mc.Comments = lineComment
            self.meta_list.append(mc)

            deleted = []
            for i, m in enumerate(meta_void):
                if m.NullCell:
                    deleted.append(i)
                    continue
                icount += 1
                m.label = icount
                update_comment(m, idLabel)
            for i in reversed(deleted):
                del meta_void[i]

            self.meta_list.extend(meta_void)

        print_warning_solids(warnSolids, warnEnclosures)

        # add plane definition to cone
        process_cones(
            self.meta_list,
            coneInfo,
            self.Surfaces,
            self.geometry_bounding_box,
            self.options,
            self.tolerances,
            self.numeric_format,
        )

        logger.info("Process finished")
        logger.info(datetime.now() - startTime)

        logger.info(f"Translation time of solid cells {tempTime1} - {tempTime0}")
        logger.info(f"Translation time of void cells {tempTime2} - {tempTime1}")

    def _decompose_solids(self, meta: bool):

        if meta:
            meta_list = self.meta_list
            description = "Decomposing solids"
        else:
            meta_list = self.enclosure_list
            description = "Decomposing enclosure solids"

        totsolid = len(meta_list)
        warningSolids = []
        for i, m in enumerate(tqdm(meta_list, desc=description)):
            if meta and m.IsEnclosure:
                continue
            logger.info(f"Decomposing solid: {i + 1}/{totsolid}")
            if self.settings.debug:
                debug_output_folder = Path("debug")
                logger.info(m.Comments)
                debug_output_folder.mkdir(parents=True, exist_ok=True)
                if m.IsEnclosure:
                    m.Solids[0].exportStep(str(debug_output_folder / f"origEnclosure_{i}.stp"))
                else:
                    m.Solids[0].exportStep(str(debug_output_folder / f"origSolid_{i}.stp"))

            comsolid, err = Decom.SplitSolid(
                Part.makeCompound(m.Solids),
                self.geometry_bounding_box,
                self.options,
                self.tolerances,
                self.numeric_format,
            )

            if err != 0:
                sus_output_folder = Path("suspicious_solids")
                sus_output_folder.mkdir(parents=True, exist_ok=True)
                if m.IsEnclosure:
                    Part.CompSolid(m.Solids).exportStep(str(sus_output_folder / f"Enclosure_original_{i}.stp"))
                    comsolid.exportStep(str(sus_output_folder / f"Enclosure_split_{i}.stp"))
                else:
                    Part.CompSolid(m.Solids).exportStep(str(sus_output_folder / f"Solid_original_{i}.stp"))
                    comsolid.exportStep(str(sus_output_folder / f"Solid_split_{i}.stp"))

                warningSolids.append(i)

            if self.settings.debug:
                if m.IsEnclosure:
                    comsolid.exportStep(str(debug_output_folder / f"compEnclosure_{i}.stp"))
                else:
                    comsolid.exportStep(str(debug_output_folder / f"compSolid_{i}.stp"))
            self.Surfaces.extend(
                Decom.extract_surfaces(
                    comsolid,
                    "All",
                    self.geometry_bounding_box,
                    self.options,
                    self.tolerances,
                    self.numeric_format,
                    MakeObj=True,
                ),
                self.options,
                self.tolerances,
                self.numeric_format,
            )
            m.set_cad_solid()
            m.update_solids(comsolid.Solids)

        return warningSolids


def update_comment(meta, idLabel):
    if meta.__commentInfo__ is None:
        return
    if meta.__commentInfo__[1] is None:
        return
    newLabel = (idLabel[i] for i in meta.__commentInfo__[1])
    meta.set_comments(void.void_comment_line((meta.__commentInfo__[0], newLabel)))


def process_cones(MetaList, coneInfo, Surfaces, UniverseBox, options, tolerances, numeric_format):
    cellId = tuple(coneInfo.keys())
    for m in MetaList:
        if m.__id__ not in cellId and not m.Void:
            continue

        if m.Void and m.__commentInfo__ is not None:
            if m.__commentInfo__[1] is None:
                continue
            cones = set()
            for Id in m.__commentInfo__[1]:
                if Id in cellId:
                    cones.update(-x for x in coneInfo[Id])
            Conv.add_cone_plane(
                m.Definition,
                cones,
                Surfaces,
                UniverseBox,
                options,
                tolerances,
                numeric_format,
            )
        elif not m.Void:
            Conv.add_cone_plane(
                m.Definition,
                coneInfo[m.__id__],
                Surfaces,
                UniverseBox,
                options,
                tolerances,
                numeric_format,
            )


def print_warning_solids(warnSolids, warnEnclosures):

    solids_logger = logging.getLogger("solids_logger")

    if warnSolids or warnEnclosures:
        pass
    else:
        return

    if warnSolids:
        lines = "Solids :\n"
        for sol in warnSolids:
            lines += "\n"
            lines += f"{sol.label}\n"
            lines += f"{sol.Comments}\n"
            lines += f"{write_mcnp_cell_def(sol.Definition)}\n"
        solids_logger.info(lines)

    if warnEnclosures:
        lines = "Enclosures :\n"
        for sol in warnEnclosures:
            lines += "\n"
            lines += f"{sol.label}\n"
            lines += f"{sol.Comments}\n"
            lines += f"{write_mcnp_cell_def(sol.Definition)}\n"

        solids_logger.info(lines)


def join_meta_lists(MList) -> typing.List[UF.GeounedSolid]:

    newMetaList = MList[0]
    if MList[0]:
        for M in MList[1:]:
            lastID = newMetaList[-1].__id__ + 1
            for i, meta in enumerate(M):
                meta.__id__ = lastID + i
                newMetaList.append(meta)

    return newMetaList


def exclude_cells(MetaList, labelList):
    voidMeta = []
    for m in MetaList:
        if m.IsEnclosure:
            continue
        found = False
        for label in labelList:
            if label in m.Comments:
                found = True
                break
        if not found:
            voidMeta.append(m)

    return voidMeta


def sort_enclosure(MetaList, meta_void, offSet=0):

    newList = {}
    for m in meta_void:
        if m.EnclosureID in newList.keys():
            newList[m.EnclosureID].append(m)
        else:
            newList[m.EnclosureID] = [m]

    icount = offSet
    idLabel = {0: 0}
    newMeta = []
    for m in MetaList:
        if m.NullCell:
            continue
        if m.IsEnclosure:
            lineComment = f"""##########################################################
             ENCLOSURE {m.EnclosureID}
##########################################################"""
            mc = UF.GeounedSolid(None)
            mc.Comments = lineComment
            newMeta.append(mc)
            for e in newList[m.EnclosureID]:
                if e.NullCell:
                    continue
                icount += 1
                e.label = icount
                idLabel[e.__id__] = e.label
                newMeta.append(e)
            lineComment = f"""##########################################################
            END  ENCLOSURE {m.EnclosureID}
##########################################################"""
            mc = UF.GeounedSolid(None)
            mc.Comments = lineComment
            newMeta.append(mc)

        else:
            icount += 1
            m.label = icount
            idLabel[m.__id__] = m.label
            newMeta.append(m)

    lineComment = """\
##########################################################
             VOID CELLS 
##########################################################"""
    mc = UF.GeounedSolid(None)
    mc.Comments = lineComment
    newMeta.append(mc)

    for v in newList[0]:
        if v.NullCell:
            continue
        icount += 1
        v.label = icount
        idLabel[v.__id__] = v.label
        newMeta.append(v)

    for m in newMeta:
        if not m.Void:
            continue
        if m.IsEnclosure:
            continue
        update_comment(m, idLabel)

    return newMeta
